# Copyright (C) 2017-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
Django production settings for swh-web.
"""

from swh.web.config import DEFAULT_CONFIG, oidc_enabled

from .common import (
    CACHES,
    DEBUG,
    INSTALLED_APPS,
    MIDDLEWARE,
    REST_FRAMEWORK,
    SECRET_KEY,
    WEBPACK_LOADER,
    swh_web_config,
)
from .common import *  # noqa

INSTALLED_APPS.append("django_minify_html")

MIDDLEWARE = MIDDLEWARE + [
    "django_minify_html.middleware.MinifyHtmlMiddleware",
]

if cache_uri := swh_web_config.get("throttling", {}).get("cache_uri"):
    memcache_settings = {
        "BACKEND": "django.core.cache.backends.memcached.PyMemcacheCache",
        "LOCATION": cache_uri,
        "OPTIONS": {
            # sets the TCP_NODELAY flag on the connection’s socket
            "no_delay": True,
            # memcache/network errors are treated as cache misses
            "ignore_exc": True,
        },
    }
    CACHES.update(
        {
            "default": memcache_settings,
            "rate-limit": memcache_settings,
        }
    )

# Setup support for proxy headers
USE_X_FORWARDED_HOST = True
SECURE_PROXY_SSL_HEADER = ("HTTP_X_FORWARDED_PROTO", "https")

# We're going through seven (or, in that case, 2) proxies thanks to Varnish
REST_FRAMEWORK["NUM_PROXIES"] = 2

db_conf = swh_web_config["production_db"]
if db_conf.get("name", "").startswith("postgresql://"):
    # poor man's support for dsn connection string...
    import psycopg

    with psycopg.connect(db_conf.get("name")) as cnx:
        db_conf["name"] = cnx.info.dbname
        db_conf["host"] = cnx.info.host
        db_conf["port"] = cnx.info.port
        db_conf["user"] = cnx.info.user
        db_conf["password"] = cnx.info.password


# https://docs.djangoproject.com/en/1.10/ref/settings/#databases
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": db_conf.get("name"),
        "HOST": db_conf.get("host"),
        "PORT": db_conf.get("port"),
        "USER": db_conf.get("user"),
        "PASSWORD": db_conf.get("password"),
    }
}

WEBPACK_LOADER["DEFAULT"]["CACHE"] = not DEBUG

if SECRET_KEY == DEFAULT_CONFIG["secret_key"][-1]:
    raise ValueError(
        "The SECRET_KEY value matches the development default, "
        "check the contents of SWH_CONFIG_FILENAME"
    )

browse_content_rate_limit = swh_web_config.get("browse_content_rate_limit", {})
RATELIMIT_ENABLE = browse_content_rate_limit.get("enabled", True)

if oidc_enabled():
    # disable django admin logs as keycloak user id overflows the
    # user id integer column in webapp database
    INSTALLED_APPS.append("django_admin_logs")
    DJANGO_ADMIN_LOGS_ENABLED = False



# Copyright (C) 2018-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from copy import deepcopy
from datetime import timedelta
import json
import os
from pathlib import Path
import random
from shutil import unpack_archive
import tempfile
import time
from typing import Dict, List, Optional, Set

from swh.core.config import merge_configs
from swh.counters import get_counters
from swh.indexer.fossology_license import FossologyLicenseIndexer
from swh.indexer.mimetype import MimetypeIndexer
from swh.indexer.storage import get_indexer_storage
from swh.indexer.storage.model import (
    OriginExtrinsicMetadataRow,
    OriginIntrinsicMetadataRow,
)
from swh.loader.git.loader import GitLoader
from swh.model.from_disk import DentryPerms
from swh.model.hashutil import DEFAULT_ALGORITHMS, hash_to_bytes, hash_to_hex
from swh.model.model import (
    Content,
    Directory,
    DirectoryEntry,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    Snapshot,
)
from swh.model.swhids import CoreSWHID, ObjectType, QualifiedSWHID
from swh.search import get_search
from swh.storage import get_storage
from swh.storage.algos.dir_iterators import dir_iterator
from swh.storage.algos.snapshot import snapshot_get_latest
from swh.storage.interface import Sha1
from swh.storage.utils import now
from swh.web import config
from swh.web.browse.utils import (
    get_mimetype_and_encoding_for_content,
    prepare_content_for_display,
    re_encode_content,
)

# Module used to initialize data that will be provided as tests input

# Base content indexer configuration
_TEST_INDEXER_BASE_CONFIG = {
    "storage": {"cls": "memory"},
    "objstorage": {
        "cls": "memory",
    },
    "indexer_storage": {
        "cls": "memory",
    },
}


def random_sha1_bytes() -> Sha1:
    return bytes(random.randint(0, 255) for _ in range(20))


def random_sha1() -> str:
    return hash_to_hex(random_sha1_bytes())


def random_sha256() -> str:
    return hash_to_hex(bytes(random.randint(0, 255) for _ in range(32)))


def random_blake2s256() -> str:
    return hash_to_hex(bytes(random.randint(0, 255) for _ in range(32)))


def random_content():
    return {
        "sha1": random_sha1(),
        "sha1_git": random_sha1(),
        "sha256": random_sha256(),
        "blake2s256": random_blake2s256(),
    }


_TEST_MIMETYPE_INDEXER_CONFIG = merge_configs(
    _TEST_INDEXER_BASE_CONFIG,
    {
        "tools": {
            "name": "file",
            "version": "1:5.30-1+deb9u1",
            "configuration": {"type": "library", "debian-package": "python3-magic"},
        }
    },
)


_TEST_LICENSE_INDEXER_CONFIG = merge_configs(
    _TEST_INDEXER_BASE_CONFIG,
    {
        "workdir": "/tmp/swh/indexer.fossology.license",
        "tools": {
            "name": "nomos",
            "version": "3.1.0rc2-31-ga2cbb8c",
            "configuration": {
                "command_line": "nomossa <filepath>",
            },
        },
    },
)


# Lightweight git repositories that will be loaded to generate
# input data for tests
ORIGIN_WITH_METADATA_FILES = "https://git.example.org/repo_with_metadata_file"
ORIGIN_WITH_CFF_FILE = "https://git.example.org/repo_with_cff_file"
ORIGIN_WITH_UNKNOWN_CODEMETA_CONTEXT = "https://example.org/codemeta/unknown/context"
ORIGIN_WITH_QUOTED_SPACE_IN_URL = "https://git.example.org/john%20doe/project"
_TEST_ORIGINS = [
    {
        "type": "git",
        "url": "https://github.com/memononen/libtess2",
        "archives": ["libtess2.zip"],
        "metadata": {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "description": (
                "Game and tools oriented refactored version of GLU tessellator."
            ),
        },
    },
    {
        "type": "git",
        "url": "https://github.com/wcoder/highlightjs-line-numbers.js",
        "archives": [
            "highlightjs-line-numbers.js.zip",
            "highlightjs-line-numbers.js_visit2.zip",
        ],
        "metadata": {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "description": "Line numbering plugin for Highlight.js",
        },
    },
    {
        "type": "git",
        "url": "https://git.example.org/repo_with_submodules",
        "archives": ["repo_with_submodules.tgz"],
        "metadata": {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "description": "This is just a sample repository with submodules",
        },
    },
    {
        "type": "git",
        "url": ORIGIN_WITH_METADATA_FILES,
        "archives": ["repo_with_metadata_file.zip"],
    },
    {
        "type": "git",
        "url": ORIGIN_WITH_CFF_FILE,
        "archives": ["repo_with_cff_file.zip"],
    },
]

_contents = {}


def _add_extra_contents(storage, contents):
    pbm_image_data = b"""P1
# PBM example
24 7
0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
0 1 1 1 1 0 0 1 1 1 1 0 0 1 1 1 1 0 0 1 1 1 1 0
0 1 0 0 0 0 0 1 0 0 0 0 0 1 0 0 0 0 0 1 0 0 1 0
0 1 1 1 0 0 0 1 1 1 0 0 0 1 1 1 0 0 0 1 1 1 1 0
0 1 0 0 0 0 0 1 0 0 0 0 0 1 0 0 0 0 0 1 0 0 0 0
0 1 0 0 0 0 0 1 1 1 1 0 0 1 1 1 1 0 0 1 0 0 0 0
0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0"""

    # add file with mimetype image/x-portable-bitmap in the archive content
    pbm_content = Content.from_data(pbm_image_data)
    storage.content_add([pbm_content])
    contents.add(pbm_content.sha1)

    # add file with mimetype application/pgp-keys in the archive content
    gpg_path = os.path.join(
        os.path.dirname(__file__), "resources/contents/other/extensions/public.gpg"
    )
    gpg_content = Content.from_data(Path(gpg_path).read_bytes())
    storage.content_add([gpg_content])
    contents.add(gpg_content.sha1)


INDEXER_TOOL = {
    "tool_name": "swh-web tests",
    "tool_version": "1.0",
    "tool_configuration": {},
}

ORIGIN_METADATA_KEY = "keywords"
ORIGIN_METADATA_VALUE = "git"

ORIGIN_MASTER_REVISION = {}
ORIGIN_MASTER_DIRECTORY = {}


class GitLoaderFromArchive(GitLoader):
    """Simple GitLoader wrapper that overrides the fetch_pack_from_origin
    operation to fetch git objects from an unpacked archive on disk instead
    of fetching them from a remote URL."""

    def __init__(self, storage, origin_url, archive_path):
        super().__init__(storage, origin_url)
        self.archive_path = archive_path

    def fetch_pack_from_origin(
        self,
        origin_url,
        base_repo,
        do_activity,
    ):
        with tempfile.TemporaryDirectory() as tmp_dir:
            unpack_archive(self.archive_path, tmp_dir)
            repo_name = os.listdir(tmp_dir)[0]
            return super().fetch_pack_from_origin(
                f"file://{tmp_dir}/{repo_name}", base_repo, do_activity
            )


def _add_origin(
    storage, search, counters, origin_url, visit_type="git", snapshot_branches={}
):
    storage.origin_add([Origin(url=origin_url)])

    counters.add("origin", [origin_url])
    date = now()
    visit = OriginVisit(origin=origin_url, date=date, type=visit_type)
    visit = storage.origin_visit_add([visit])[0]
    counters.add("origin_visit", [f"{visit.unique_key()}"])
    if snapshot_branches is not None:
        snapshot = Snapshot.from_dict({"branches": snapshot_branches})
        storage.snapshot_add([snapshot])
        counters.add("snapshot", [snapshot.id])
    visit_status = OriginVisitStatus(
        origin=origin_url,
        visit=visit.visit,
        date=date + timedelta(minutes=1),
        type=visit.type,
        status="full",
        snapshot=snapshot.id if snapshot_branches is not None else None,
    )
    storage.origin_visit_status_add([visit_status])
    counters.add("origin_visit_status", [f"{visit_status.unique_key()}"])
    search.origin_update(
        [
            {
                "url": origin_url,
                "has_visits": True,
                "visit_types": [visit_type],
                "snapshot_id": (
                    snapshot.id.hex() if snapshot_branches is not None else None
                ),
            }
        ]
    )


def _add_origin_with_unknown_codemeta_context(storage, search, counters):
    codemeta = {
        "@context": "https://example.org/codemeta/3.0",
        "author": {"name": "Jane Doe"},
        "name": "Example Software",
        "url": "http://example.org/",
        "datePublished": "2023-10-10",
    }

    content = Content.from_data(json.dumps(codemeta).encode())
    storage.content_add([content])

    directory = Directory(
        entries=(
            DirectoryEntry(
                name=b"codemeta.json",
                type="file",
                target=content.sha1_git,
                perms=DentryPerms.content,
            ),
        )
    )
    storage.directory_add([directory])

    _add_origin(
        storage,
        search,
        counters,
        origin_url=ORIGIN_WITH_UNKNOWN_CODEMETA_CONTEXT,
        visit_type="tar",
        snapshot_branches={
            b"HEAD": {"target_type": "directory", "target": directory.id}
        },
    )


# Tests data initialization
def _init_tests_data():
    # To hold reference to the memory storage
    storage = get_storage("memory")

    # Create search instance
    search = get_search("memory")
    search.initialize()
    search.origin_update({"url": origin["url"]} for origin in _TEST_ORIGINS)

    # create the counters instance
    counters = get_counters("memory")

    # Create indexer storage instance that will be shared by indexers
    idx_storage = get_indexer_storage("memory")

    # Declare a test tool for origin intrinsic metadata tests
    idx_tool = idx_storage.indexer_configuration_add([INDEXER_TOOL])[0]
    INDEXER_TOOL["id"] = idx_tool["id"]

    # Load git repositories from archives
    for origin in _TEST_ORIGINS:
        for i, archive_ in enumerate(origin["archives"]):
            if i > 0:
                # ensure visit dates will be different when simulating
                # multiple visits of an origin
                time.sleep(1)
            origin_repo_archive = os.path.join(
                os.path.dirname(__file__), "resources/repos/%s" % archive_
            )
            loader = GitLoaderFromArchive(
                storage,
                origin["url"],
                archive_path=origin_repo_archive,
            )
            result = loader.load()
            assert result["status"] == "eventful"

        ori = storage.origin_get([origin["url"]])[0]
        origin.update(ori.to_dict())  # add an 'id' key if enabled
        search.origin_update(
            [
                {
                    "url": origin["url"],
                    "has_visits": True,
                    "visit_types": ["git"],
                    "snapshot_id": loader.snapshot.id.hex(),
                }
            ]
        )

    for i in range(250):
        _add_origin(
            storage,
            search,
            counters,
            origin_url=f"https://many.origins/{i + 1}",
            visit_type="tar",
        )

    sha1s: Set[Sha1] = set()
    directories = set()
    revisions = set()
    releases = set()
    snapshots = set()
    swhids = []

    content_path = {}

    # Get all objects loaded into the test archive
    common_metadata = {ORIGIN_METADATA_KEY: ORIGIN_METADATA_VALUE}
    for origin in _TEST_ORIGINS:
        origin_revisions = set()
        snp = snapshot_get_latest(storage, origin["url"])
        swhids.append(
            QualifiedSWHID(
                object_type=ObjectType.SNAPSHOT, object_id=snp.id, origin=origin["url"]
            )
        )
        snapshots.add(hash_to_hex(snp.id))
        for branch_name, branch_data in snp.branches.items():
            target_type = branch_data.target_type.value
            if target_type == "revision":
                origin_revisions.add(branch_data.target)
                swhids.append(
                    QualifiedSWHID(
                        object_type=ObjectType.REVISION,
                        object_id=branch_data.target,
                        origin=origin["url"],
                        visit=CoreSWHID(
                            object_type=ObjectType.SNAPSHOT, object_id=snp.id
                        ),
                    )
                )
                if b"master" in branch_name:
                    # Add some origin intrinsic and extrinsic metadata for tests
                    metadata = common_metadata
                    metadata.update(origin.get("metadata", {}))
                    revision = storage.revision_get([branch_data.target])[0]
                    origin_intrinsic_metadata = OriginIntrinsicMetadataRow(
                        id=origin["url"],
                        from_directory=revision.directory,
                        indexer_configuration_id=idx_tool["id"],
                        metadata=metadata,
                        mappings=[],
                    )
                    idx_storage.origin_intrinsic_metadata_add(
                        [origin_intrinsic_metadata]
                    )
                    origin_extrinsic_metadata = OriginExtrinsicMetadataRow(
                        id=origin["url"],
                        from_remd_id=revision.directory,
                        indexer_configuration_id=idx_tool["id"],
                        metadata=metadata,
                        mappings=[],
                    )
                    idx_storage.origin_extrinsic_metadata_add(
                        [origin_extrinsic_metadata]
                    )
                    search.origin_update([{"url": origin["url"], "jsonld": metadata}])

                    ORIGIN_MASTER_REVISION[origin["url"]] = hash_to_hex(
                        branch_data.target
                    )
                    ORIGIN_MASTER_DIRECTORY[origin["url"]] = hash_to_hex(
                        revision.directory
                    )
            elif target_type == "release":
                release = storage.release_get([branch_data.target])[0]
                counters.add("person", (release.author.fullname,))
                origin_revisions.add(release.target)
                releases.add(hash_to_hex(branch_data.target))
                swhids.append(
                    QualifiedSWHID(
                        object_type=ObjectType.RELEASE,
                        object_id=branch_data.target,
                        origin=origin["url"],
                        visit=CoreSWHID(
                            object_type=ObjectType.SNAPSHOT, object_id=snp.id
                        ),
                    )
                )

        for rev in storage.revision_log(origin_revisions):
            counters.add(
                "person", (rev["author"]["fullname"], rev["committer"]["fullname"])
            )
            rev_id = rev["id"]
            revisions.add(rev_id)

        for rev in storage.revision_get(sorted(origin_revisions)):
            if rev is None:
                continue
            dir_id = rev.directory
            directories.add(hash_to_hex(dir_id))
            for entry in dir_iterator(storage, dir_id):
                if entry["type"] == "file":
                    sha1s.add(entry["sha1"])
                    content_path[entry["sha1"]] = "/".join(
                        [hash_to_hex(dir_id), entry["path"].decode("utf-8")]
                    )
                    swhids.append(
                        QualifiedSWHID(
                            object_type=ObjectType.CONTENT,
                            object_id=entry["sha1_git"],
                            origin=origin["url"],
                            visit=CoreSWHID(
                                object_type=ObjectType.SNAPSHOT, object_id=snp.id
                            ),
                            anchor=CoreSWHID(
                                object_type=ObjectType.REVISION, object_id=rev.id
                            ),
                            path=b"/" + entry["path"],
                        )
                    )
                elif entry["type"] == "dir":
                    directories.add(hash_to_hex(entry["target"]))
                    swhids.append(
                        QualifiedSWHID(
                            object_type=ObjectType.DIRECTORY,
                            object_id=entry["target"],
                            origin=origin["url"],
                            visit=CoreSWHID(
                                object_type=ObjectType.SNAPSHOT, object_id=snp.id
                            ),
                            anchor=CoreSWHID(
                                object_type=ObjectType.REVISION, object_id=rev.id
                            ),
                            path=b"/" + entry["path"] + b"/",
                        )
                    )

    _add_extra_contents(storage, sha1s)

    # Get all checksums for each content
    result: List[Optional[Content]] = storage.content_get(list(sha1s))

    contents: List[Dict] = []
    for content in result:
        assert content is not None
        sha1 = hash_to_hex(content.sha1)
        content_metadata = {
            algo: hash_to_hex(getattr(content, algo)) for algo in DEFAULT_ALGORITHMS
        }

        path = ""
        if content.sha1 in content_path:
            path = content_path[content.sha1]

        cnt_data = storage.content_get_data({"sha1": content.sha1})
        assert cnt_data is not None
        mimetype, encoding = get_mimetype_and_encoding_for_content(cnt_data)
        _, _, cnt_data = re_encode_content(mimetype, encoding, cnt_data)

        content_display_data = prepare_content_for_display(
            cnt_data, mimetype, encoding, path
        )

        content_metadata.update(
            {
                "path": path,
                "mimetype": mimetype,
                "encoding": encoding,
                "hljs_language": content_display_data["language"],
                "raw_data": cnt_data,
                "data": content_display_data["content_data"],
            }
        )

        _contents[sha1] = content_metadata
        contents.append(content_metadata)

    # Add the empty directory to the test archive
    storage.directory_add([Directory(entries=())])

    # Add empty content to the test archive
    storage.content_add([Content.from_data(data=b"")])

    # origins used in cypress tests for origins search
    _add_origin(
        storage,
        search,
        counters,
        origin_url="https://example.org/project/download.php?version=2.0",
        visit_type="tar",
        snapshot_branches={
            b"refs/heads/master": {
                "target_type": "revision",
                "target": next(iter(revisions)),
            }
        },
    )

    _add_origin(
        storage,
        search,
        counters,
        origin_url="https://example.org/project/with/null/snapshot",
        visit_type="hg",
        snapshot_branches=None,
    )

    _add_origin(
        storage,
        search,
        counters,
        origin_url=ORIGIN_WITH_QUOTED_SPACE_IN_URL,
        visit_type="git",
    )

    _add_origin_with_unknown_codemeta_context(storage, search, counters)

    # Add fake git origin with pull request branches
    _add_origin(
        storage,
        search,
        counters,
        origin_url="https://git.example.org/project",
        snapshot_branches={
            b"refs/heads/master": {
                "target_type": "revision",
                "target": next(iter(revisions)),
            },
            **{
                f"refs/pull/{i}".encode(): {
                    "target_type": "revision",
                    "target": next(iter(revisions)),
                }
                for i in range(300)
            },
        },
    )

    snapshots_list = list(sorted(snapshots))

    # Add fake origin with multiple visit types
    origin = Origin(url="https://example.org/project/multiple/visit/types")
    storage.origin_add([origin])
    visit_types = ["git", "git-checkout"]
    for i, visit_type in enumerate(visit_types):
        origin_visit = storage.origin_visit_add(
            [
                OriginVisit(
                    origin=origin.url,
                    date=now() + timedelta(days=i),
                    type=visit_type,
                )
            ]
        )[0]
        visit_status = OriginVisitStatus(
            origin=origin.url,
            visit=origin_visit.visit,
            date=now() + timedelta(days=i),
            status="full",
            snapshot=hash_to_bytes(snapshots_list[i]),
        )
        storage.origin_visit_status_add([visit_status])
    search.origin_update(
        [
            {
                "url": origin.url,
                "visit_types": visit_types,
                "has_visits": True,
                "snapshot_id": snapshots_list[i],
            }
        ]
    )

    counters.add("revision", revisions)
    counters.add("release", releases)
    counters.add("directory", directories)
    counters.add("content", [content["sha1"] for content in contents])

    # Return tests data
    return {
        "search": search,
        "storage": storage,
        "idx_storage": idx_storage,
        "counters": counters,
        "origins": _TEST_ORIGINS,
        "contents": list(sorted(contents, key=lambda c: c["sha1"])),
        "directories": list(sorted(directories)),
        "releases": list(sorted(releases)),
        "revisions": list(sorted(map(hash_to_hex, revisions))),
        "snapshots": snapshots_list,
        "swhids": swhids,
    }


def _init_indexers(tests_data):
    # Instantiate content indexers that will be used in tests
    # and force them to use the memory storages
    indexers = {}
    for idx_name, idx_class, idx_config in (
        ("mimetype_indexer", MimetypeIndexer, _TEST_MIMETYPE_INDEXER_CONFIG),
        ("license_indexer", FossologyLicenseIndexer, _TEST_LICENSE_INDEXER_CONFIG),
    ):
        idx = idx_class(config=idx_config)
        idx.storage = tests_data["storage"]
        idx.objstorage = tests_data["storage"].objstorage
        idx.idx_storage = tests_data["idx_storage"]
        idx.register_tools(idx.config["tools"])
        indexers[idx_name] = idx

    return indexers


def get_content(content_sha1):
    return _contents.get(content_sha1)


_tests_data = None
_current_tests_data = None
_indexer_loggers = {}


def get_tests_data(reset=False):
    """
    Initialize tests data and return them in a dict.
    """
    global _tests_data, _current_tests_data
    if _tests_data is None:
        _tests_data = _init_tests_data()
        indexers = _init_indexers(_tests_data)
        for name, idx in indexers.items():
            # pytest makes the loggers use a temporary file; and deepcopy
            # requires serializability. So we remove them, and add them
            # back after the copy.
            _indexer_loggers[name] = idx.log
            del idx.log
        _tests_data.update(indexers)
    if reset or _current_tests_data is None:
        _current_tests_data = deepcopy(_tests_data)
        for name, logger in _indexer_loggers.items():
            _current_tests_data[name].log = logger
    return _current_tests_data


def override_storages(storage, idx_storage, search, counters):
    """
    Helper function to replace the storages from which archive data
    are fetched.
    """
    swh_config = config.get_config()
    swh_config.update(
        {
            "storage": storage,
            "indexer_storage": idx_storage,
            "search": search,
            "counters": counters,
        }
    )



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.urls.converters import REGISTERED_CONVERTERS, register_converter

from swh.model.swhids import CoreSWHID, ExtendedSWHID, QualifiedSWHID


class SWHIDConverter:
    """Simple SWHID django url path converter to simplify view URLs registration.

    When parsing URL, this simply wraps the SWHID regexp and pass the matched
    string as django view parameter, it is up to the view to verify the SWHID
    validity and returns a 400 status code if it is not.

    When reversing to an URL, only SWHID objects and str are allowed.
    """

    regex = "(swh|SWH):[0-9]+:[A-Za-z]+:[0-9A-Fa-f]+(;.*)?"

    def to_python(self, value):
        value_split = value.split(";", maxsplit=1)
        if len(value_split) > 1:
            return f"{value_split[0].lower()};{value_split[1]}"
        else:
            return value.lower()

    def to_url(self, value):
        if isinstance(value, (CoreSWHID, ExtendedSWHID, QualifiedSWHID, str)):
            return str(value)
        raise ValueError()


def register_url_path_converters() -> None:
    """Adds :class:`.SWHIDConverter` to django's url path converters.

    This method is called by the root `URLconf` but also in every app's `URLconf` that
    needs it to ease testing, that's why we need to check if it is already registered
    before adding it.
    """
    if "swhid" not in REGISTERED_CONVERTERS:
        register_converter(SWHIDConverter, "swhid")


register_url_path_converters()



# Copyright (C) 2020-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import random

import pytest

from swh.model.hashutil import hash_to_bytes
from swh.model.model import (
    Release,
    ReleaseTargetType,
    Snapshot,
    SnapshotBranch,
    SnapshotTargetType,
)
from swh.model.swhids import ObjectType
from swh.web.browse.snapshot_context import (
    _get_release,
    get_origin_visit_snapshot,
    get_snapshot_content,
    get_snapshot_context,
)
from swh.web.browse.utils import gen_revision_url
from swh.web.utils import format_utc_iso_date, reverse
from swh.web.utils.identifiers import gen_swhid
from swh.web.utils.origin_visits import get_origin_visit, get_origin_visits
from swh.web.utils.typing import (
    SnapshotBranchInfo,
    SnapshotContext,
    SnapshotReleaseInfo,
)


def test_get_origin_visit_snapshot_simple(archive_data, origin_with_multiple_visits):
    visits = archive_data.origin_visit_get(origin_with_multiple_visits["url"])

    for visit in visits:
        snapshot = archive_data.snapshot_get(visit["snapshot"])
        branches = []
        releases = []

        def _process_branch_data(branch, branch_data, alias=False):
            if branch_data["target_type"] == "revision":
                rev_data = archive_data.revision_get(branch_data["target"])
                branches.append(
                    SnapshotBranchInfo(
                        name=branch,
                        alias=alias,
                        target_type="revision",
                        target=branch_data["target"],
                        directory=rev_data["directory"],
                        date=format_utc_iso_date(rev_data["date"]),
                        message=None,
                        url=None,
                    )
                )
            elif branch_data["target_type"] == "release":
                rel_data = archive_data.release_get(branch_data["target"])
                rev_data = archive_data.revision_get(rel_data["target"])
                releases.append(
                    SnapshotReleaseInfo(
                        name=rel_data["name"],
                        alias=alias,
                        branch_name=branch,
                        date=format_utc_iso_date(rel_data["date"]),
                        id=rel_data["id"],
                        message=None,
                        target_type=rel_data["target_type"],
                        target=rel_data["target"],
                        directory=rev_data["directory"],
                        url=None,
                    )
                )

        aliases = {}

        for branch in sorted(snapshot["branches"].keys()):
            branch_data = snapshot["branches"][branch]
            if branch_data["target_type"] == "alias":
                target_data = snapshot["branches"][branch_data["target"]]
                aliases[branch] = target_data
                _process_branch_data(branch, target_data, alias=True)
            else:
                _process_branch_data(branch, branch_data)

        assert branches and releases, "Incomplete test data."

        origin_visit_branches = get_origin_visit_snapshot(
            origin_with_multiple_visits, visit_id=visit["visit"]
        )

        releases = list(reversed(sorted(releases, key=lambda r: r["name"])))

        assert origin_visit_branches == (branches, releases, aliases)


def test_get_snapshot_context_no_origin(archive_data, snapshot):
    for browse_context, kwargs in (
        ("content", {"snapshot_id": snapshot, "path": "/some/path"}),
        ("directory", {"snapshot_id": snapshot}),
        ("log", {"snapshot_id": snapshot}),
    ):
        url_args = {"snapshot_id": snapshot}

        query_params = dict(kwargs)
        query_params["visit_type"] = None
        query_params.pop("snapshot_id")

        snapshot_context = get_snapshot_context(**kwargs, browse_context=browse_context)

        branches, releases, _ = get_snapshot_content(snapshot)

        revision_id = None
        root_directory = None
        for branch in branches:
            if branch["name"] == "HEAD":
                revision_id = branch["target"]
                root_directory = branch["directory"]
            branch["url"] = reverse(
                f"browse-snapshot-{browse_context}",
                url_args=url_args,
                query_params={"branch": branch["name"], **query_params},
            )
        for release in releases:
            release["url"] = reverse(
                f"browse-snapshot-{browse_context}",
                url_args=url_args,
                query_params={"release": release["name"], **query_params},
            )

        branches_url = reverse("browse-snapshot-branches", url_args=url_args)
        releases_url = reverse("browse-snapshot-releases", url_args=url_args)
        directory_url = reverse("browse-snapshot-directory", url_args=url_args)
        is_empty = not branches and not releases
        snapshot_swhid = gen_swhid(ObjectType.SNAPSHOT, snapshot)
        snapshot_sizes = archive_data.snapshot_count_branches(snapshot)

        expected = SnapshotContext(
            branch="HEAD",
            branch_alias=True,
            branches=branches,
            branches_url=branches_url,
            is_empty=is_empty,
            origin_info=None,
            origin_visits_url=None,
            release=None,
            release_alias=False,
            release_id=None,
            query_params=query_params,
            releases=releases,
            releases_url=releases_url,
            revision_id=revision_id,
            revision_info=_get_revision_info(archive_data, revision_id),
            root_directory=root_directory,
            snapshot_id=snapshot,
            snapshot_sizes=snapshot_sizes,
            snapshot_swhid=snapshot_swhid,
            url_args=url_args,
            visit_info=None,
            browse_url=directory_url,
        )

        if revision_id:
            expected["revision_info"]["revision_url"] = gen_revision_url(
                revision_id, snapshot_context
            )

        assert snapshot_context == expected

        _check_branch_release_revision_parameters(
            archive_data, expected, browse_context, kwargs, branches, releases
        )


@pytest.mark.xfail  # flaky test
def test_get_snapshot_context_with_origin(archive_data, origin_with_multiple_visits):
    origin_visits = get_origin_visits(origin_with_multiple_visits["url"])

    timestamp = format_utc_iso_date(origin_visits[0]["date"], "%Y-%m-%dT%H:%M:%SZ")
    visit_id = origin_visits[1]["visit"]

    origin_url = origin_with_multiple_visits["url"]

    for browse_context, kwargs in (
        ("content", {"origin_url": origin_url, "path": "/some/path"}),
        ("directory", {"origin_url": origin_url}),
        ("log", {"origin_url": origin_url}),
        (
            "directory",
            {
                "origin_url": origin_url,
                "timestamp": timestamp,
            },
        ),
        (
            "directory",
            {
                "origin_url": origin_url,
                "visit_id": visit_id,
            },
        ),
    ):
        visit_id = kwargs["visit_id"] if "visit_id" in kwargs else None
        visit_ts = kwargs["timestamp"] if "timestamp" in kwargs else None
        visit_info = get_origin_visit(
            kwargs["origin_url"], visit_ts=visit_ts, visit_id=visit_id
        )

        snapshot = visit_info["snapshot"]

        snapshot_context = get_snapshot_context(**kwargs, browse_context=browse_context)

        query_params = dict(kwargs)
        query_params["visit_type"] = None

        branches, releases, _ = get_snapshot_content(snapshot)
        revision_id = None
        root_directory = None
        for branch in branches:
            if branch["name"] == "HEAD":
                revision_id = branch["target"]
                root_directory = branch["directory"]
            branch["url"] = reverse(
                f"browse-origin-{browse_context}",
                query_params={"branch": branch["name"], **query_params},
            )
        for release in releases:
            release["url"] = reverse(
                f"browse-origin-{browse_context}",
                query_params={"release": release["name"], **query_params},
            )

        query_params.pop("path", None)

        branches_url = reverse("browse-origin-branches", query_params=query_params)
        releases_url = reverse("browse-origin-releases", query_params=query_params)
        origin_visits_url = reverse(
            "browse-origin-visits", query_params={"origin_url": kwargs["origin_url"]}
        )
        is_empty = not branches and not releases
        snapshot_swhid = gen_swhid(ObjectType.SNAPSHOT, snapshot)
        snapshot_sizes = archive_data.snapshot_count_branches(snapshot)

        visit_info["url"] = directory_url = reverse(
            "browse-origin-directory", query_params=query_params
        )
        visit_info["formatted_date"] = format_utc_iso_date(visit_info["date"])

        if "path" in kwargs:
            query_params["path"] = kwargs["path"]

        expected = SnapshotContext(
            branch="HEAD",
            branch_alias=True,
            branches=branches,
            branches_url=branches_url,
            is_empty=is_empty,
            origin_info={"url": origin_url, "visit_types": ["git"]},
            origin_visits_url=origin_visits_url,
            release=None,
            release_alias=False,
            release_id=None,
            query_params=query_params,
            releases=releases,
            releases_url=releases_url,
            revision_id=revision_id,
            revision_info=_get_revision_info(archive_data, revision_id),
            root_directory=root_directory,
            snapshot_id=snapshot,
            snapshot_sizes=snapshot_sizes,
            snapshot_swhid=snapshot_swhid,
            url_args={},
            visit_info=visit_info,
            browse_url=directory_url,
        )

        if revision_id:
            expected["revision_info"]["revision_url"] = gen_revision_url(
                revision_id, snapshot_context
            )
        assert snapshot_context == expected

        _check_branch_release_revision_parameters(
            archive_data, expected, browse_context, kwargs, branches, releases
        )


def _check_branch_release_revision_parameters(
    archive_data,
    base_expected_context,
    browse_context,
    kwargs,
    branches,
    releases,
):
    branch = random.choice(branches)

    snapshot_context = get_snapshot_context(
        **kwargs, browse_context=browse_context, branch_name=branch["name"]
    )

    url_args = dict(kwargs)
    url_args.pop("path", None)
    url_args.pop("timestamp", None)
    url_args.pop("visit_id", None)
    url_args.pop("origin_url", None)

    query_params = dict(kwargs)
    query_params.pop("snapshot_id", None)
    query_params["visit_type"] = None

    expected_branch = dict(base_expected_context)
    expected_branch["branch"] = branch["name"]
    expected_branch["branch_alias"] = branch["alias"]
    expected_branch["revision_id"] = branch["target"]
    expected_branch["revision_info"] = _get_revision_info(
        archive_data, branch["target"]
    )
    expected_branch["root_directory"] = branch["directory"]
    expected_branch["query_params"] = {"branch": branch["name"], **query_params}
    expected_branch["revision_info"]["revision_url"] = gen_revision_url(
        branch["target"], expected_branch
    )

    assert snapshot_context == expected_branch

    if releases:
        release = random.choice(releases)

        snapshot_context = get_snapshot_context(
            **kwargs, browse_context=browse_context, release_name=release["name"]
        )

        expected_release = dict(base_expected_context)
        expected_release["branch"] = None
        expected_release["branch_alias"] = False
        expected_release["release"] = release["name"]
        expected_release["release_id"] = release["id"]
        if release["target_type"] == "revision":
            expected_release["revision_id"] = release["target"]
            expected_release["revision_info"] = _get_revision_info(
                archive_data, release["target"]
            )
        expected_release["root_directory"] = release["directory"]
        expected_release["query_params"] = {"release": release["name"], **query_params}
        expected_release["revision_info"]["revision_url"] = gen_revision_url(
            release["target"], expected_release
        )

        assert snapshot_context == expected_release

    revision_log = archive_data.revision_log(branch["target"])
    revision = revision_log[-1]

    snapshot_context = get_snapshot_context(
        **kwargs, browse_context=browse_context, revision_id=revision["id"]
    )

    if "origin_url" in kwargs:
        view_name = f"browse-origin-{browse_context}"
    else:
        view_name = f"browse-snapshot-{browse_context}"

    kwargs.pop("visit_id", None)

    revision_browse_url = reverse(
        view_name,
        url_args=url_args,
        query_params={"revision": revision["id"], **query_params},
    )

    branches.append(
        SnapshotBranchInfo(
            name=revision["id"],
            alias=False,
            target_type="revision",
            target=revision["id"],
            directory=revision["directory"],
            date=revision["date"],
            message=revision["message"],
            url=revision_browse_url,
        )
    )

    expected_revision = dict(base_expected_context)
    expected_revision["branch"] = None
    expected_revision["branch_alias"] = False
    expected_revision["branches"] = branches
    expected_revision["revision_id"] = revision["id"]
    expected_revision["revision_info"] = _get_revision_info(
        archive_data, revision["id"]
    )
    expected_revision["root_directory"] = revision["directory"]
    expected_revision["query_params"] = {"revision": revision["id"], **query_params}
    expected_revision["revision_info"]["revision_url"] = gen_revision_url(
        revision["id"], expected_revision
    )

    assert snapshot_context == expected_revision


def test_get_release_large_snapshot(archive_data, origin_with_releases):
    snapshot = archive_data.snapshot_get_latest(origin_with_releases["url"])
    release_id = random.choice(
        [
            v["target"]
            for v in snapshot["branches"].values()
            if v["target_type"] == "release"
        ]
    )
    release_data = archive_data.release_get(release_id)
    # simulate large snapshot processing by providing releases parameter
    # as an empty list
    release = _get_release(
        releases=[], release_name=release_data["name"], snapshot_id=snapshot["id"]
    )

    assert release_data["name"] == release["name"]
    assert release_data["id"] == release["id"]

    # check mercurial/package release branch edge cases
    for branch_prefix in ("tags", "releases"):
        snapshot = Snapshot(
            branches={
                f"{branch_prefix}/{release['name']}".encode(): SnapshotBranch(
                    target=hash_to_bytes(release_id),
                    target_type=SnapshotTargetType.RELEASE,
                ),
            }
        )
        archive_data.snapshot_add([snapshot])
        release = _get_release(
            releases=[],
            release_name=release_data["name"],
            snapshot_id=snapshot.id.hex(),
        )
        assert release_data["name"] == release["name"]
        assert release_data["id"] == release["id"]


def _get_revision_info(archive_data, revision_id):
    revision_info = None
    if revision_id:
        revision_info = archive_data.revision_get(revision_id)
        revision_info["message_header"] = revision_info["message"].split("\n")[0]
        revision_info["date"] = format_utc_iso_date(revision_info["date"])
        revision_info["committer_date"] = format_utc_iso_date(
            revision_info["committer_date"]
        )
    return revision_info


def test_get_snapshot_context_revision_release(archive_data, revision):
    release_name = "v1.0.0"
    release = Release(
        name=release_name.encode(),
        message=f"release {release_name}".encode(),
        target=hash_to_bytes(revision),
        target_type=ReleaseTargetType.REVISION,
        synthetic=True,
    )
    archive_data.release_add([release])

    snapshot = Snapshot(
        branches={
            b"HEAD": SnapshotBranch(
                target=release_name.encode(), target_type=SnapshotTargetType.ALIAS
            ),
            release_name.encode(): SnapshotBranch(
                target=release.id, target_type=SnapshotTargetType.RELEASE
            ),
        },
    )
    archive_data.snapshot_add([snapshot])

    snapshot_no_head = Snapshot(
        branches={
            release_name.encode(): SnapshotBranch(
                target=release.id, target_type=SnapshotTargetType.RELEASE
            ),
        },
    )
    archive_data.snapshot_add([snapshot_no_head])

    revision_data = archive_data.revision_get(revision)

    for params in (
        {"snapshot_id": snapshot.id.hex()},
        {"snapshot_id": snapshot.id.hex(), "release_name": release_name},
        {"snapshot_id": snapshot_no_head.id.hex()},
    ):
        snapshot_context = get_snapshot_context(**params)

        assert snapshot_context["branches"] == []
        assert snapshot_context["releases"] != []
        assert snapshot_context["release"] == release_name
        assert snapshot_context["release_id"] == release.id.hex()
        assert snapshot_context["revision_id"] == revision
        assert snapshot_context["root_directory"] == revision_data["directory"]


def test_get_snapshot_context_directory_release(archive_data, directory):
    release_name = "v1.0.0"
    release = Release(
        name=release_name.encode(),
        message=f"release {release_name}".encode(),
        target=hash_to_bytes(directory),
        target_type=ReleaseTargetType.DIRECTORY,
        synthetic=True,
    )
    archive_data.release_add([release])

    snapshot = Snapshot(
        branches={
            b"HEAD": SnapshotBranch(
                target=release_name.encode(), target_type=SnapshotTargetType.ALIAS
            ),
            release_name.encode(): SnapshotBranch(
                target=release.id, target_type=SnapshotTargetType.RELEASE
            ),
        },
    )
    archive_data.snapshot_add([snapshot])

    snapshot_no_head = Snapshot(
        branches={
            release_name.encode(): SnapshotBranch(
                target=release.id, target_type=SnapshotTargetType.RELEASE
            ),
        },
    )
    archive_data.snapshot_add([snapshot_no_head])

    for params in (
        {"snapshot_id": snapshot.id.hex()},
        {"snapshot_id": snapshot.id.hex(), "release_name": release_name},
        {"snapshot_id": snapshot_no_head.id.hex()},
    ):
        snapshot_context = get_snapshot_context(**params)

        assert snapshot_context["branches"] == []
        assert snapshot_context["releases"] != []
        assert snapshot_context["release"] == release_name
        assert snapshot_context["release_id"] == release.id.hex()
        assert snapshot_context["revision_id"] is None
        assert snapshot_context["root_directory"] == directory



# Copyright (C) 2022-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime

import pytest

from django.utils.timezone import get_current_timezone

from swh.web.auth.utils import (
    ADD_FORGE_MODERATOR_PERMISSION,
    ADMIN_LIST_DEPOSIT_PERMISSION,
    API_PROVENANCE_PERMISSION,
    API_SAVE_BULK_PERMISSION,
    API_SAVE_ORIGIN_PERMISSION,
    MAILMAP_ADMIN_PERMISSION,
    MAILMAP_PERMISSION,
    SWH_AMBASSADOR_PERMISSION,
)

APP_NAME = "swh_web_auth"

MIGRATION_0005 = "0005_usermailmapevent"
MIGRATION_0006 = "0006_fix_mailmap_admin_user_id"
MIGRATION_0007 = "0007_mailmap_django_app"
MIGRATION_0008 = "0008_create_webapp_permissions"
MIGRATION_0009 = "0009_create_provenance_permission"


def test_fix_mailmap_admin_user_id(migrator):
    state = migrator.apply_tested_migration((APP_NAME, MIGRATION_0005))
    UserMailmap = state.apps.get_model(APP_NAME, "UserMailmap")
    user_id = "45"

    UserMailmap.objects.create(
        user_id=user_id,
        from_email="user@example.org",
        from_email_verified=True,
        display_name="New display name",
    )

    UserMailmap.objects.filter(user_id=user_id).update(
        last_update_date=datetime(
            2022, 2, 11, 14, 16, 13, 614000, get_current_timezone()
        )
    )

    assert UserMailmap.objects.filter(user_id=user_id).count() == 1
    assert UserMailmap.objects.filter(user_id=None).count() == 0

    state = migrator.apply_tested_migration((APP_NAME, MIGRATION_0006))
    UserMailmap = state.apps.get_model(APP_NAME, "UserMailmap")

    assert UserMailmap.objects.filter(user_id=user_id).count() == 0
    assert UserMailmap.objects.filter(user_id=None).count() == 1


def test_mailmap_django_app(migrator):
    state = migrator.apply_tested_migration((APP_NAME, MIGRATION_0006))
    UserMailmap = state.apps.get_model(APP_NAME, "UserMailmap")
    assert UserMailmap

    # UserMailmap model moved to swh_web_mailmap django application
    state = migrator.apply_tested_migration((APP_NAME, MIGRATION_0007))
    with pytest.raises(
        LookupError, match="App 'swh_web_auth' doesn't have a 'UserMailmap' model."
    ):
        state.apps.get_model(APP_NAME, "UserMailmap")


@pytest.mark.skip(
    reason=(
        "FIXME: looks like permissions are created before"
        "0008_create_webapp_permissions"
    )
)
def test_create_webapp_permissions(migrator):
    state = migrator.apply_initial_migration((APP_NAME, MIGRATION_0007))
    Permission = state.apps.get_model("auth", "Permission")
    webapp_permissions = [
        SWH_AMBASSADOR_PERMISSION,
        API_SAVE_ORIGIN_PERMISSION,
        ADMIN_LIST_DEPOSIT_PERMISSION,
        MAILMAP_PERMISSION,
        ADD_FORGE_MODERATOR_PERMISSION,
        MAILMAP_ADMIN_PERMISSION,
        API_SAVE_BULK_PERMISSION,
    ]
    for webapp_permission in webapp_permissions:
        assert Permission.objects.filter(name=webapp_permission).count() == 0

    state = migrator.apply_tested_migration((APP_NAME, MIGRATION_0008))
    Permission = state.apps.get_model("auth", "Permission")
    for webapp_permission in webapp_permissions:
        assert Permission.objects.filter(name=webapp_permission).count() == 1


def test_create_provenance_permission(migrator):
    state = migrator.apply_initial_migration((APP_NAME, MIGRATION_0008))
    Permission = state.apps.get_model("auth", "Permission")
    assert Permission.objects.filter(name=API_PROVENANCE_PERMISSION).count() == 0
    state = migrator.apply_tested_migration((APP_NAME, MIGRATION_0009))
    Permission = state.apps.get_model("auth", "Permission")
    assert Permission.objects.filter(name=API_PROVENANCE_PERMISSION).count() == 1



# Copyright (C) 2015-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Optional

from rest_framework.request import Request

from swh.web.api import utils
from swh.web.api.apidoc import api_doc, format_docstring
from swh.web.api.apiurls import api_route
from swh.web.api.views.utils import api_lookup
from swh.web.utils import archive


@api_route(
    r"/directory/(?P<sha1_git>[0-9a-f]+)/",
    "api-1-directory",
    checksum_args=["sha1_git"],
)
@api_route(
    r"/directory/(?P<sha1_git>[0-9a-f]+)/(?P<path>.+)/",
    "api-1-directory",
    checksum_args=["sha1_git"],
)
@api_doc("/directory/", category="Archive")
@format_docstring()
def api_directory(request: Request, sha1_git: str, path: Optional[str] = None):
    """
    .. http:get:: /api/1/directory/(sha1_git)/[(path)/]

        Get information about directory objects.
        Directories are identified by **sha1** checksums, compatible with Git
        directory identifiers.
        See :func:`swh.model.git_objects.directory_git_object` in our data model
        module for details about how they are computed.

        When given only a directory identifier, this endpoint returns information
        about the directory itself, returning its content (usually a list of directory
        entries). When given a directory identifier and a path, this endpoint returns
        information about the directory entry pointed by the relative path, starting
        path resolution from the given directory.

        :param string sha1_git: hexadecimal representation of the directory
            **sha1_git** identifier
        :param string path: optional parameter to get information about the
            directory entry pointed by that relative path

        {common_headers}

        :>jsonarr object checksums: object holding the computed checksum values for
            a directory entry (only for file entries)
        :>jsonarr string dir_id: **sha1_git** identifier of the requested directory
        :>jsonarr number length: length of a directory entry in bytes (only for file
            entries) for getting information about the content MIME type
        :>jsonarr string name: the directory entry name
        :>jsonarr number perms: permissions for the directory entry
        :>jsonarr string target: **sha1_git** identifier of the directory entry
        :>jsonarr string target_url: link to
            :http:get:`/api/1/content/[(hash_type):](hash)/`
            or :http:get:`/api/1/directory/(sha1_git)/[(path)/]`
            depending on the directory entry type
        :>jsonarr string type: the type of the directory entry, can be either
            ``dir``, ``file`` or ``rev``

        :statuscode 200: no error
        :statuscode 400: an invalid **hash_type** or **hash** has been provided
        :statuscode 404: requested directory cannot be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`directory/977fc4b98c0e85816348cebd3b12026407c368b6/`
    """
    if path:
        error_msg_path = (
            "Entry with path %s relative to directory " "with sha1_git %s not found."
        ) % (path, sha1_git)
        return api_lookup(
            archive.lookup_directory_with_path,
            sha1_git,
            path,
            notfound_msg=error_msg_path,
            enrich_fn=utils.enrich_directory_entry,
            request=request,
        )
    else:
        error_msg_nopath = "Directory with sha1_git %s not found." % sha1_git
        return api_lookup(
            archive.lookup_directory,
            sha1_git,
            notfound_msg=error_msg_nopath,
            enrich_fn=utils.enrich_directory_entry,
            request=request,
        )



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest
from pytest_django.asserts import (
    assertContains,
    assertNotContains,
    assertRedirects,
    assertTemplateUsed,
)

from django.conf import settings
from django.core.paginator import Page
from django.forms.models import model_to_dict as dj_model_to_dict
from django.utils import timezone

from swh.web.alter.forms import (
    INITIALS_REASONS,
    AlterationAccessForm,
    AlterationAdminForm,
    AlterationForm,
    AlterationSearchForm,
    ConfirmationForm,
    EmailVerificationForm,
    MessageAdminForm,
    MessageForm,
    OriginAdminForm,
    OriginSearchForm,
    OriginSelectForm,
    ReasonsForm,
)
from swh.web.alter.models import (
    Alteration,
    AlterationCategory,
    AlterationStatus,
    EventRecipient,
    OriginOutcome,
    OriginOwnership,
    Token,
)
from swh.web.alter.utils import (
    SESSION_ALTERATION_IDS,
    SESSION_VERIFIED_EMAIL,
    has_access,
    verified_email,
)
from swh.web.tests.helpers import check_html_get_response, check_html_post_response
from swh.web.utils import reverse


@pytest.fixture
def _session_step0(client, db):
    session = client.session
    session[SESSION_VERIFIED_EMAIL] = "mh@m.it"
    session.save()


@pytest.fixture
def _session_step1(client, _session_step0):
    session = client.session
    session["alteration_category"] = AlterationCategory.COPYRIGHT
    session.save()


@pytest.fixture
def _session_step2(client, _session_step1):
    session = client.session
    session["alteration_origins"] = [
        "http://example.org/username/django",
        "http://example.org/username/flask",
    ]
    session.save()


@pytest.fixture
def _session_step3(client, _session_step2):
    session = client.session
    session["alteration_reasons"] = "My reasons"
    session["alteration_expected_outcome"] = "My expected outcome"
    session.save()


@pytest.fixture
def _session_access(client, alteration):
    session = client.session
    session[SESSION_ALTERATION_IDS] = [str(alteration.pk)]
    session.save()


@pytest.fixture
def admin_client(client, alter_admin):
    client.force_login(alter_admin)
    return client


def model_to_dict(instance, fields=None, exclude=None):
    """Return a dict containing the data in ``instance``

    This calls django's `model_to_dict` and then removes None values from the dict to
    avoid TypeError when POSTing data with the TestClient.
    """
    data = dj_model_to_dict(instance, fields, exclude)
    return {k: v for k, v in data.items() if v is not None}


def test_content_policies_unverified_email(client):
    response = check_html_get_response(
        client,
        reverse("content-policy"),
        status_code=200,
        template_used="content_policies.html",
    )
    assert response.context["next_step"] == "alteration-email"
    assertContains(response, reverse("alteration-email"))


@pytest.mark.django_db
def test_content_policies_verified_email(client):
    session = client.session
    session[SESSION_VERIFIED_EMAIL] = "test@mail.local"
    session.save()
    response = check_html_get_response(
        client,
        reverse("content-policy"),
        status_code=200,
        template_used="content_policies.html",
    )
    assert response.context["next_step"] == "alteration-category"
    assertContains(response, reverse("alteration-category"))


def test_email(client):
    response = check_html_get_response(
        client,
        reverse("alteration-email"),
        status_code=200,
        template_used="assistant_email.html",
    )
    assert isinstance(response.context["form"], EmailVerificationForm)
    assert response.context["process"]["email"].active
    assertTemplateUsed(response, "includes/steps.html")


@pytest.mark.django_db
def test_email_post(client, mailoutbox):
    data = {"email": "test@mail.local"}
    response = check_html_post_response(
        client,
        reverse("alteration-email"),
        status_code=302,
        data=data,
    )
    assertRedirects(response, reverse("alteration-email"))
    assert len(mailoutbox) == 1
    message = mailoutbox[0]
    assert message.subject == "Please confirm your email address"


@pytest.mark.django_db
def test_email_verification(client):
    token = Token.create_for("test@mail.local")
    response = check_html_get_response(
        client,
        token.get_absolute_url(),
        status_code=302,
    )
    assertRedirects(response, reverse("alteration-category"))
    assert verified_email(response.wsgi_request) == "test@mail.local"


@pytest.mark.django_db
def test_email_verification_expired(client):
    token = Token.create_for("test@mail.local")
    token.expires_at = timezone.now()
    token.save()
    response = check_html_get_response(
        client,
        token.get_absolute_url(),
        status_code=302,
    )
    assertRedirects(response, reverse("alteration-email"))
    assert not verified_email(response.wsgi_request)


def test_category_missing_email(client):
    response = client.get(reverse("alteration-category"))
    assertRedirects(response, reverse("alteration-email"))


def test_category(client, _session_step0):
    response = check_html_get_response(
        client,
        reverse("alteration-category"),
        status_code=200,
        template_used="assistant_category.html",
    )
    assert response.context["process"]["category"].active
    assertTemplateUsed(response, "includes/steps.html")


def test_category_post(client, _session_step0):
    response = check_html_post_response(
        client,
        reverse("alteration-category"),
        status_code=302,
        data={"category": AlterationCategory.COPYRIGHT},
    )
    assert response["location"] == reverse("alteration-origins")
    assert client.session["alteration_category"] == AlterationCategory.COPYRIGHT


def test_origins_missing_category(client, _session_step0):
    response = check_html_get_response(
        client, reverse("alteration-origins"), status_code=302
    )
    assert response["location"] == reverse("alteration-category")


def test_origins_context(client, _session_step1):
    context = check_html_get_response(
        client, reverse("alteration-origins"), status_code=200
    ).context
    assert isinstance(context["search_form"], OriginSearchForm)
    assert isinstance(context["origins_form"], OriginSelectForm)
    assert context["process"]["origins"].active
    assert not context["results"]
    assert not context["query"]


def test_origins_search(client, mocker, _session_step1):
    url = reverse("alteration-origins", query_params={"query": "username"})
    mocker.patch(
        "swh.web.alter.views.search_origin",
        return_value=(["http://example.org/username/repo"], None, None),
    )
    response = check_html_get_response(
        client,
        url,
        status_code=200,
        template_used="assistant_origins.html",
    )
    assert response.context["results"] == ["http://example.org/username/repo"]


def test_origins_post(client, origin, _session_step1):
    response = check_html_post_response(
        client,
        reverse("alteration-origins"),
        status_code=302,
        data={"urls": [origin["url"]]},
    )
    assertRedirects(response, reverse("alteration-reasons"))
    assert client.session["alteration_origins"] == [origin["url"]]


def test_reasons_without_origins(client, _session_step1):
    response = check_html_get_response(
        client, reverse("alteration-reasons"), status_code=302
    )
    assert response["location"] == reverse("alteration-origins")


def test_reasons(client, _session_step2):
    response = check_html_get_response(
        client,
        reverse("alteration-reasons"),
        status_code=200,
        template_used="assistant_reasons.html",
    )
    form = response.context["form"]
    assert isinstance(form, ReasonsForm)
    # form fields are pre-filed with text depending on the Alteration category
    assert (
        form.initial["reasons"]
        == INITIALS_REASONS[AlterationCategory.COPYRIGHT]["reasons"]
    )
    assert (
        form.initial["expected_outcome"]
        == INITIALS_REASONS[AlterationCategory.COPYRIGHT]["expected_outcome"]
    )
    assert response.context["process"]["reasons"].active
    assertTemplateUsed(response, "includes/steps.html")


def test_reasons_post(client, _session_step2):
    response = check_html_post_response(
        client,
        reverse("alteration-reasons"),
        status_code=302,
        data={"reasons": "My reasons", "expected_outcome": "My expected outcome"},
    )
    assertRedirects(response, reverse("alteration-summary"))
    assert client.session["alteration_reasons"] == "My reasons"
    assert client.session["alteration_expected_outcome"] == "My expected outcome"


def test_summary(client, _session_step3):
    response = check_html_get_response(
        client,
        reverse("alteration-summary"),
        status_code=200,
        template_used="assistant_summary.html",
    )
    form = response.context["form"]
    assert isinstance(form, ConfirmationForm)
    # all session values must be found in the summary
    values = [v for v in client.session.values() if not isinstance(v, list)]
    values += client.session["alteration_origins"]
    for value in values:
        assertContains(response, value)
    assert response.context["process"]["summary"].active
    assertTemplateUsed(response, "includes/steps.html")


def test_summary_without_reasons(client, _session_step2):
    response = client.get(reverse("alteration-summary"))
    assertRedirects(response, reverse("alteration-reasons"))


@pytest.mark.django_db
def test_summary_post(client, _session_step3):
    response = check_html_post_response(
        client,
        reverse("alteration-summary"),
        status_code=302,
        data={"confirm": True},
    )
    # An Alteration request has been created
    alteration = Alteration.objects.get(email="mh@m.it")
    assertRedirects(response, alteration.get_absolute_url())
    # a session var has been added to authorize current browser
    has_access(response.wsgi_request, alteration.pk)
    # Other session vars have been deleted
    remaining_keys = [
        k
        for k in client.session.keys()
        if k.startswith("alteration_")
        and k != SESSION_ALTERATION_IDS
        and k != SESSION_VERIFIED_EMAIL
    ]
    assert not remaining_keys


def test_details_is_protected(client, alteration):
    response = check_html_get_response(
        client,
        alteration.get_absolute_url(),
        status_code=302,
    )
    assertRedirects(response, reverse("alteration-access", {"pk": alteration.pk}))


def test_details(client, alteration, _session_access):
    response = check_html_get_response(
        client,
        alteration.get_absolute_url(),
        status_code=200,
        template_used="alteration_details.html",
    )
    context = response.context
    assert context["alteration"] == alteration
    assert len(context["events"]) == 1  # only one public event
    assert isinstance(context["message_form"], MessageForm)
    assert isinstance(context["alteration_form"], AlterationForm)
    for event in alteration.events.all():
        if event.internal:
            assertNotContains(response, event.content)
        else:
            assertContains(response, event.content)


def test_details_post(client, alteration, _session_access):
    data = model_to_dict(alteration)
    data["reasons"] = "My updated reasons"
    response = check_html_post_response(
        client,
        alteration.get_absolute_url(),
        status_code=302,
        data=data,
    )
    assertRedirects(response, alteration.get_absolute_url())
    alteration.refresh_from_db()
    assert alteration.reasons == "My updated reasons"
    event = alteration.events.first()
    assert event.author == "Requester"
    assert "reasons" in event.content


def test_message_protected(client, alteration):
    data = {"content": "Please block my whole namespace"}
    response = check_html_post_response(
        client,
        reverse("alteration-message", {"pk": alteration.pk}),
        status_code=302,
        data=data,
    )
    assertRedirects(response, reverse("alteration-access", {"pk": alteration.pk}))


def test_message_post(client, alteration, _session_access, mailoutbox):
    data = {"content": "Please block my whole namespace"}
    response = check_html_post_response(
        client,
        reverse("alteration-message", {"pk": alteration.pk}),
        status_code=302,
        data=data,
    )
    assertRedirects(response, alteration.get_absolute_url())
    event = alteration.events.first()
    assert event.content == "Please block my whole namespace"
    assert len(mailoutbox) == 1
    message = mailoutbox[0]
    assert message.subject == f"New message on {alteration}"


def test_access(client, alteration):
    response = check_html_get_response(
        client,
        reverse("alteration-access", {"pk": alteration.pk}),
        status_code=200,
        template_used="alteration_access.html",
    )
    assert isinstance(response.context["form"], AlterationAccessForm)


def test_access_post(client, alteration, mailoutbox):
    url = reverse("alteration-access", {"pk": alteration.pk})
    response = check_html_post_response(
        client, url, status_code=302, data={"email": alteration.email}
    )
    assertRedirects(response, url)
    assert len(mailoutbox) == 1
    message = mailoutbox[0]
    assert message.subject == "Access to your alteration request"


def test_access_post_wrong_email(client, alteration, mailoutbox):
    url = reverse("alteration-access", {"pk": alteration.pk})
    response = check_html_post_response(
        client, url, status_code=302, data={"email": f"wrong.{alteration.email}"}
    )
    assertRedirects(response, url)
    assert len(mailoutbox) == 0


def test_access_link(client, token):
    alteration = token.alteration
    response = check_html_get_response(
        client,
        token.get_absolute_url(),
        status_code=302,
    )
    assertRedirects(response, alteration.get_absolute_url())
    assert has_access(response.wsgi_request, alteration.pk)


def test_access_link_expired(client, alteration, token):
    token.expires_at = timezone.now()
    token.save()
    response = check_html_get_response(
        client,
        token.get_absolute_url(),
        status_code=302,
    )
    assertRedirects(response, reverse("alteration-access", {"pk": alteration.pk}))


def test_admin_dashboard_protected(client):
    response = check_html_get_response(
        client,
        reverse("alteration-dashboard"),
        status_code=302,
    )
    assert reverse(settings.LOGIN_URL) in response.headers["Location"]


@pytest.mark.django_db
def test_admin_dashboard(admin_client, alteration):
    response = check_html_get_response(
        admin_client,
        reverse("alteration-dashboard"),
        status_code=200,
        template_used="admin_dashboard.html",
    )
    context = response.context
    assert isinstance(context["form"], AlterationSearchForm)
    assert isinstance(context["page"], Page)
    assert alteration in context["page"]
    assertContains(response, str(alteration))


@pytest.mark.django_db
@pytest.mark.parametrize(
    "query,status,expected",
    [
        ("ZZZ", None, False),
        ("example.com", None, True),
        (None, AlterationStatus.VALIDATING, True),
    ],
)
def test_admin_dashboard_search(admin_client, alteration, query, status, expected):
    response = check_html_get_response(
        admin_client,
        reverse(
            "alteration-dashboard", query_params={"query": query, "status": status}
        ),
        status_code=200,
        template_used="admin_dashboard.html",
    )
    page = response.context["page"]
    if expected:
        assert alteration in page
    else:
        assert alteration not in page


def test_admin_alteration_protected(client, alteration):
    response = check_html_get_response(
        client,
        alteration.get_admin_url(),
        status_code=302,
    )
    assert reverse(settings.LOGIN_URL) in response.headers["Location"]


@pytest.mark.django_db
def test_admin_alteration(admin_client, alteration):
    response = check_html_get_response(
        admin_client,
        alteration.get_admin_url(),
        status_code=200,
        template_used="admin_alteration.html",
    )
    context = response.context
    assert context["alteration"] == alteration
    assert isinstance(context["alteration_form"], AlterationAdminForm)
    assert isinstance(context["message_form"], MessageAdminForm)
    assert isinstance(context["origin_create_form"], OriginAdminForm)
    assert len(context["origin_forms"]) == alteration.origins.count()
    origins = set()
    for form in context["origin_forms"]:
        origins.add(form.instance)
        assert isinstance(form, OriginAdminForm)
    assert origins == set(alteration.origins.all())
    for event in alteration.events.all():
        assertContains(response, event.content)


@pytest.mark.django_db
def test_admin_alteration_post(admin_client, alteration):
    data = model_to_dict(alteration)
    data["expected_outcome"] = "Takedown, block + mailmap"
    response = check_html_post_response(
        admin_client, alteration.get_admin_url(), status_code=302, data=data
    )
    assertRedirects(response, alteration.get_admin_url())
    alteration.refresh_from_db()
    assert alteration.expected_outcome == "Takedown, block + mailmap"


def test_admin_origin_protected(client, alteration):
    origin = alteration.origins.first()
    url = reverse(
        "alteration-origin-admin", {"alteration_pk": alteration.pk, "pk": origin.pk}
    )
    response = check_html_post_response(
        client, url, status_code=302, data=model_to_dict(origin)
    )
    assert reverse(settings.LOGIN_URL) in response.headers["Location"]


@pytest.mark.django_db
def test_admin_origin_post(admin_client, alteration):
    origin = alteration.origins.first()
    url = reverse(
        "alteration-origin-admin", {"alteration_pk": alteration.pk, "pk": origin.pk}
    )
    data = model_to_dict(origin)
    data["code_license"] = "CeCILL"
    response = check_html_post_response(admin_client, url, status_code=302, data=data)
    assertRedirects(response, alteration.get_admin_url())
    origin.refresh_from_db()
    assert origin.code_license == "CeCILL"


@pytest.mark.django_db
def test_admin_origin_create_post(admin_client, alteration, origin):
    url = reverse("alteration-origin-admin-create", {"alteration_pk": alteration.pk})
    data = {
        "url": origin["url"],
        "available": True,
        "outcome": OriginOutcome.VALIDATING,
        "ownership": OriginOwnership.UNKNOWN,
    }
    response = check_html_post_response(admin_client, url, status_code=302, data=data)
    assertRedirects(response, alteration.get_admin_url())
    assert alteration.origins.get(url=origin["url"]).available


def test_admin_message_protected(client, alteration):
    url = reverse("alteration-message-admin", {"pk": alteration.pk})
    data = {
        "recipient": EventRecipient.MANAGER,
        "internal": True,
        "content": "Please check this alteration request",
    }
    response = check_html_post_response(client, url, status_code=302, data=data)
    assert reverse(settings.LOGIN_URL) in response.headers["Location"]


@pytest.mark.django_db
def test_admin_message_post(admin_client, alter_admin, alteration):
    url = reverse("alteration-message-admin", {"pk": alteration.pk})
    data = {
        "recipient": EventRecipient.MANAGER,
        "internal": True,
        "content": "Please check this alteration request",
    }
    response = check_html_post_response(admin_client, url, status_code=302, data=data)
    assertRedirects(response, alteration.get_admin_url())
    event = alteration.events.first()
    assert event.author == alter_admin.username
    assert event.content == "Please check this alteration request"


def test_event_admin_protected(client, alteration):
    url = reverse(
        "alteration-event-admin",
        {"alteration_pk": alteration.pk, "pk": alteration.events.first().pk},
    )
    data = {"content": "Updated content"}
    response = check_html_post_response(client, url, status_code=302, data=data)
    assert reverse(settings.LOGIN_URL) in response.headers["Location"]


@pytest.mark.django_db
def test_event_admin_post(admin_client, alter_admin, alteration):
    url = reverse(
        "alteration-event-admin",
        {"alteration_pk": alteration.pk, "pk": alteration.events.first().pk},
    )
    data = {"content": "Updated content"}
    response = check_html_post_response(admin_client, url, status_code=302, data=data)
    assertRedirects(response, alteration.get_admin_url())
    event = alteration.events.first()
    assert event.content == "Updated content"



# Copyright (C) 2020-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
from typing import Any, Dict, Optional, Union, cast

from cryptography.fernet import InvalidToken

from django.contrib.auth.decorators import login_required
from django.core.paginator import Paginator
from django.http import HttpRequest
from django.http.response import (
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseForbidden,
    HttpResponseRedirect,
    JsonResponse,
)
from django.shortcuts import render
from django.views.decorators.http import require_http_methods

from swh.auth.django.models import OIDCUser
from swh.auth.django.utils import keycloak_oidc_client
from swh.auth.django.views import get_oidc_login_data, oidc_login_view
from swh.auth.keycloak import KeycloakError, keycloak_error_message
from swh.web.auth.models import OIDCUserOfflineTokens
from swh.web.auth.utils import decrypt_data, encrypt_data
from swh.web.config import get_config
from swh.web.utils import datatables_pagination_params, reverse
from swh.web.utils.exc import ForbiddenExc


def oidc_generate_bearer_token(request: HttpRequest) -> HttpResponse:
    if not request.user.is_authenticated or not isinstance(request.user, OIDCUser):
        return HttpResponseForbidden()
    redirect_uri = reverse("oidc-generate-bearer-token-complete", request=request)
    return oidc_login_view(
        request, redirect_uri=redirect_uri, scope="openid offline_access"
    )


def oidc_generate_bearer_token_complete(request: HttpRequest) -> HttpResponse:
    if not request.user.is_authenticated or not isinstance(request.user, OIDCUser):
        raise ForbiddenExc("You are not allowed to generate bearer tokens.")
    if "error" in request.GET:
        raise Exception(request.GET["error"])

    login_data = get_oidc_login_data(request)
    oidc_client = keycloak_oidc_client()
    oidc_profile = oidc_client.authorization_code(
        code=request.GET["code"],
        code_verifier=login_data["code_verifier"],
        redirect_uri=login_data["redirect_uri"],
    )
    user = cast(OIDCUser, request.user)
    token = oidc_profile["refresh_token"]
    secret = get_config()["secret_key"].encode()
    salt = user.sub.encode()
    encrypted_token = encrypt_data(token.encode(), secret, salt)
    OIDCUserOfflineTokens.objects.create(
        user_id=str(user.id), offline_token=encrypted_token
    ).save()
    return HttpResponseRedirect(reverse("oidc-profile") + "#tokens")


def oidc_list_bearer_tokens(request: HttpRequest) -> HttpResponse:
    if not request.user.is_authenticated or not isinstance(request.user, OIDCUser):
        return HttpResponseForbidden()

    tokens = OIDCUserOfflineTokens.objects.filter(user_id=str(request.user.id))
    tokens = tokens.order_by("-creation_date")

    length, page = datatables_pagination_params(request)

    paginator = Paginator(tokens, length)

    tokens_data = [
        {"id": t.id, "creation_date": t.creation_date.isoformat()}
        for t in paginator.page(int(page)).object_list
    ]

    table_data: Dict[str, Any] = {}
    table_data["recordsTotal"] = len(tokens_data)
    table_data["draw"] = int(request.GET.get("draw", 1))
    table_data["data"] = tokens_data
    table_data["recordsFiltered"] = len(tokens_data)
    return JsonResponse(table_data)


def _encrypted_token_bytes(token: Union[bytes, memoryview]) -> bytes:
    # token has been retrieved from a PostgreSQL database
    if isinstance(token, memoryview):
        return token.tobytes()
    else:
        return token


def _decrypt_user_token(
    token_data: OIDCUserOfflineTokens, salt: bytes
) -> Optional[bytes]:
    current_secret = get_config()["secret_key"].encode()
    fallback_secrets = [
        secret.encode() for secret in get_config()["secret_key_fallbacks"]
    ]
    decrypted_token = None
    for secret in [current_secret] + fallback_secrets:
        try:
            decrypted_token = decrypt_data(
                _encrypted_token_bytes(token_data.offline_token), secret, salt
            )
            if secret in fallback_secrets:
                # re-encrypt user token if django secret got rotated
                token_data.offline_token = encrypt_data(
                    decrypted_token, current_secret, salt
                )
                token_data.save()
            break
        except InvalidToken:
            pass
    return decrypted_token


@require_http_methods(["POST"])
def oidc_get_bearer_token(request: HttpRequest) -> HttpResponse:
    if not request.user.is_authenticated or not isinstance(request.user, OIDCUser):
        return HttpResponseForbidden()
    try:
        data = json.loads(request.body.decode("ascii"))
        user = cast(OIDCUser, request.user)
        token_data = OIDCUserOfflineTokens.objects.get(id=data["token_id"])
        decrypted_token = _decrypt_user_token(token_data, salt=user.sub.encode())
        if decrypted_token is None:
            return HttpResponseBadRequest(
                "Stored bearer token could not be decrypted, please generate a new one.",
                content_type="text/plain",
            )
        refresh_token = decrypted_token.decode("ascii")
        # check token is still valid
        oidc_client = keycloak_oidc_client()
        oidc_client.refresh_token(refresh_token)
        return HttpResponse(refresh_token, content_type="text/plain")
    except KeycloakError as ke:
        error_msg = keycloak_error_message(ke)
        if error_msg in (
            "invalid_grant: Offline session not active",
            "invalid_grant: Offline user session not found",
        ):
            error_msg = "Bearer token has expired, please generate a new one."
        return HttpResponseBadRequest(error_msg, content_type="text/plain")


@require_http_methods(["POST"])
def oidc_revoke_bearer_tokens(request: HttpRequest) -> HttpResponse:
    if not request.user.is_authenticated or not isinstance(request.user, OIDCUser):
        return HttpResponseForbidden()

    data = json.loads(request.body.decode("ascii"))
    user = cast(OIDCUser, request.user)
    for token_id in data["token_ids"]:
        token_data = OIDCUserOfflineTokens.objects.get(id=token_id)
        decrypted_token = _decrypt_user_token(token_data, salt=user.sub.encode())
        if decrypted_token is not None:
            oidc_client = keycloak_oidc_client()
            oidc_client.logout(decrypted_token.decode("ascii"))
        token_data.delete()
    return HttpResponse(status=200)


@login_required(login_url="oidc-login")
def oidc_profile_view(request: HttpRequest) -> HttpResponse:
    return render(request, "profile.html")



# Copyright (C) 2022-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
from typing import Any, Dict, List

from django.contrib.auth.decorators import user_passes_test
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.core.paginator import Paginator
from django.db.models import Q
from django.http.request import HttpRequest
from django.http.response import HttpResponse, HttpResponseForbidden, JsonResponse
from django.shortcuts import render

from swh.web.add_forge_now.api_views import (
    AddForgeNowRequestActorRole,
    AddForgeNowRequestForm,
    AddForgeNowRequestHistory,
    AddForgeNowRequestPublicSerializer,
    AddForgeNowRequestSerializer,
    AddForgeNowRequestStatus,
)
from swh.web.add_forge_now.models import Request as AddForgeRequest
from swh.web.add_forge_now.models import RequestHistory
from swh.web.auth.utils import is_add_forge_now_moderator
from swh.web.utils import datatables_order_params, datatables_pagination_params


def add_forge_request_list_datatables(request: HttpRequest) -> HttpResponse:
    """Dedicated endpoint used by datatables to display the add-forge
    requests in the Web UI.
    """

    draw = int(request.GET.get("draw", 0))

    add_forge_requests = AddForgeRequest.objects.all()

    table_data: Dict[str, Any] = {
        "recordsTotal": add_forge_requests.count(),
        "draw": draw,
    }

    search_value = request.GET.get("search[value]")

    field_order = datatables_order_params(request, "id", "desc")

    add_forge_requests = add_forge_requests.order_by(*field_order)

    per_page, page_num = datatables_pagination_params(request)

    if search_value:
        add_forge_requests = add_forge_requests.filter(
            Q(forge_type__icontains=search_value)
            | Q(forge_url__icontains=search_value)
            | Q(status__icontains=search_value)
        )

    if (
        int(request.GET.get("user_requests_only", "0"))
        and request.user.is_authenticated
    ):
        add_forge_requests = add_forge_requests.filter(
            submitter_name=request.user.username
        )

    paginator = Paginator(add_forge_requests, per_page)
    page = paginator.page(page_num)

    if is_add_forge_now_moderator(request.user):
        requests = AddForgeNowRequestSerializer(page.object_list, many=True).data
    else:
        requests = AddForgeNowRequestPublicSerializer(page.object_list, many=True).data

    results = [dict(req) for req in requests]
    table_data["recordsFiltered"] = add_forge_requests.count()
    table_data["data"] = results
    return JsonResponse(table_data)


FORGE_TYPES: List[str] = [
    "bitbucket",
    "cgit",
    "forgejo",
    "gitea",
    "gitiles",
    "gitlab",
    "gitweb",
    "gogs",
    "heptapod",
    "stagit",
]


def create_request_create(request):
    """View to create a new 'add_forge_now' request."""

    submit_status = ""
    alert_level = ""
    status_code = 200

    if request.method == "POST":
        if not request.user.is_authenticated:
            return HttpResponseForbidden(
                "You must be authenticated to update a new add-forge request"
            )
        add_forge_request = AddForgeRequest()
        form = AddForgeNowRequestForm(request.POST, instance=add_forge_request)

        if form.errors:
            status_code = 400
            alert_level = "danger"
            submit_status = json.dumps(form.errors)
        else:
            try:
                existing_request = AddForgeRequest.objects.get(
                    forge_url=add_forge_request.forge_url
                )
            except ObjectDoesNotExist:
                pass
            else:
                status_code = 409
                submit_status = (
                    f"Request for forge already exists (id {existing_request.id})"
                )
                alert_level = "danger"

            if not submit_status:
                assert isinstance(request.user, User)
                add_forge_request.submitter_name = request.user.username
                add_forge_request.submitter_email = request.user.email

                form.save()

                request_history = AddForgeNowRequestHistory()
                request_history.request = add_forge_request
                request_history.new_status = AddForgeNowRequestStatus.PENDING.name
                request_history.actor = request.user.username
                request_history.actor_role = AddForgeNowRequestActorRole.SUBMITTER.name
                request_history.save()

                add_forge_request.last_modified_date = request_history.date
                add_forge_request.save()

                submit_status = "Your request has been submitted"
                alert_level = "success"

    return render(
        request,
        "add-forge-creation-form.html",
        {
            "forge_types": FORGE_TYPES,
            "submit_status": submit_status,
            "alert_level": alert_level,
        },
        status=status_code,
    )


def create_request_list(request):
    """View to list existing 'add_forge_now' requests."""

    user_requests = []

    if request.user.is_authenticated and "no_js" in request.GET:
        user_requests = AddForgeRequest.objects.filter(
            submitter_name=request.user.username
        ).order_by("-submission_date")

    return render(
        request,
        "add-forge-list.html",
        {
            "user_requests": [
                AddForgeNowRequestPublicSerializer(user_request).data
                for user_request in user_requests
            ]
        },
    )


def create_request_help(request):
    """View to explain 'add_forge_now'."""

    return render(
        request,
        "add-forge-help.html",
    )


@user_passes_test(is_add_forge_now_moderator)
def create_request_message_source(request: HttpRequest, id: int) -> HttpResponse:
    """View to retrieve the message source for a given request history entry"""

    try:
        history_entry = RequestHistory.objects.select_related("request").get(
            pk=id, message_source__isnull=False
        )
        assert history_entry.message_source is not None
    except RequestHistory.DoesNotExist:
        return HttpResponse(status=404)

    response = HttpResponse(
        bytes(history_entry.message_source), content_type="text/email"
    )
    filename = f"add-forge-now-{history_entry.request.forge_domain}-message{id}.eml"

    response["Content-Disposition"] = f'attachment; filename="{filename}"'

    return response



# Copyright (C) 2018-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Optional, cast

from django_ratelimit.decorators import ratelimit

from django.core.paginator import Paginator
from django.db.models import Q
from django.http import JsonResponse
from django.shortcuts import render
from django.utils.safestring import mark_safe

from swh.web.auth.utils import SWH_AMBASSADOR_PERMISSION, privileged_user
from swh.web.save_code_now.models import SaveOriginRequest
from swh.web.save_code_now.origin_save import (
    create_save_origin_request,
    get_savable_visit_types,
    get_save_origin_task_info,
)
from swh.web.utils import datatables_order_params, datatables_pagination_params, reverse


@ratelimit(key="user_or_ip", rate="60/m", method="POST", block=False)
def origin_save_help_view(request):
    visit_type = request.POST.get("visit_type")
    origin_url = request.POST.get("origin_url")
    submit_status = ""
    alert_level = ""
    if request.method == "POST":
        if request.limited:
            alert_level = "danger"
            submit_status = (
                'The rate limit for "save code now" requests has been reached. '
                "Please try again later."
            )

        else:
            try:
                sor = create_save_origin_request(
                    visit_type,
                    origin_url,
                    privileged_user=request.user.is_staff,
                    user_id=cast(Optional[int], request.user.id),
                )
            except Exception as exc:
                alert_level = "danger"
                submit_status = (
                    f'An error happened when submitting the "save code now request": '
                    f"{exc}"
                )
            else:
                request_url = reverse(
                    "api-1-save-origin", url_args={"request_id": sor["id"]}
                )
                if sor["save_request_status"] == "accepted":
                    alert_level = "success"
                    submit_status = (
                        'The "save code now" request has been accepted and will be '
                        "processed as soon as possible. You can follow its processing "
                        f'by querying <a href="{request_url}">{request_url}</a>.'
                    )

                else:
                    alert_level = "warning"
                    submit_status = (
                        'The "save code now" request has been put in pending state and '
                        "may be accepted for processing after manual review. You can "
                        "follow its processing by querying "
                        f'<a href="{request_url}">{request_url}</a>.'
                    )
    else:
        origin_url = request.GET.get("origin_url", "")
    return render(
        request,
        "origin-save-help.html",
        {
            "heading": ("Request the saving of a software origin into the archive"),
            "visit_types": get_savable_visit_types(
                privileged_user(request, permissions=[SWH_AMBASSADOR_PERMISSION])
            ),
            "origin_url": origin_url,
            "visit_type": visit_type,
            "submit_status": mark_safe(submit_status),
            "alert_level": alert_level,
        },
    )


def origin_save_list_view(request):
    return render(
        request,
        "origin-save-list.html",
        {
            "heading": ("Request the saving of a software origin into the archive"),
            "visit_types": get_savable_visit_types(
                privileged_user(request, permissions=[SWH_AMBASSADOR_PERMISSION])
            ),
            "origin_url": request.GET.get("origin_url", ""),
            "submit_status": "",
        },
    )


def origin_save_requests_list(request, status):
    if status != "all":
        save_requests = SaveOriginRequest.objects.filter(status=status)
    else:
        save_requests = SaveOriginRequest.objects.all()

    table_data = {}
    table_data["recordsTotal"] = save_requests.count()
    table_data["draw"] = int(request.GET.get("draw", 1))

    search_value = request.GET.get("search[value]")

    field_order = datatables_order_params(request, "id", "desc")

    save_requests = save_requests.order_by(*field_order)

    length, page = datatables_pagination_params(request)

    if search_value:
        save_requests = save_requests.filter(
            Q(status__icontains=search_value)
            | Q(loading_task_status__icontains=search_value)
            | Q(visit_type__icontains=search_value)
            | Q(origin_url__icontains=search_value)
        )

    if (
        int(request.GET.get("user_requests_only", "0"))
        and request.user.is_authenticated
    ):
        save_requests = save_requests.filter(user_ids__contains=f'"{request.user.id}"')

    table_data["recordsFiltered"] = save_requests.count()
    paginator = Paginator(save_requests, length)
    table_data["data"] = [sor.to_dict() for sor in paginator.page(page).object_list]
    return JsonResponse(table_data)


def save_origin_task_info(request, save_request_id):
    request_info = get_save_origin_task_info(save_request_id)
    for date_field in ("scheduled", "started", "ended"):
        if date_field in request_info and request_info[date_field] is not None:
            request_info[date_field] = request_info[date_field].isoformat()
    return JsonResponse(request_info)



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import csv
import io
from typing import Any, Dict, List, Optional, Set, Tuple, TypedDict
from urllib.parse import urlparse
from uuid import UUID

from django.core.paginator import EmptyPage, Paginator
from django.utils.encoding import force_str
from rest_framework.decorators import parser_classes
from rest_framework.exceptions import ParseError
from rest_framework.parsers import BaseParser, JSONParser
from rest_framework.request import Request
from rest_framework.response import Response

from swh.model.swhids import CoreSWHID, ObjectType, QualifiedSWHID
from swh.scheduler.utils import create_oneshot_task
from swh.web.api.apidoc import api_doc, format_docstring
from swh.web.api.apiurls import APIUrls, api_route
from swh.web.api.parsers import YAMLParser
from swh.web.auth.utils import API_SAVE_BULK_PERMISSION
from swh.web.config import get_config, scheduler
from swh.web.save_bulk.models import SaveBulkOrigin, SaveBulkRequest
from swh.web.save_code_now.origin_save import validate_origin_url
from swh.web.utils import datetime_to_utc, reverse
from swh.web.utils.exc import BadInputExc, ForbiddenExc, NotFoundExc, UnauthorizedExc

save_bulk_api_urls = APIUrls()

SUPPORTED_VISIT_TYPES = {"bzr", "cvs", "hg", "git", "svn", "tarball-directory"}


def _register_request_and_origins_in_db(
    user_id: str, origins: Set[Tuple[str, str]]
) -> str:
    save_bulk_request = SaveBulkRequest.objects.create(user_id=user_id)

    # create new submitted origins in database
    save_bulk_origins = SaveBulkOrigin.objects.bulk_create(
        [
            SaveBulkOrigin(origin_url=origin_url, visit_type=visit_type)
            for origin_url, visit_type in origins
        ],
        update_conflicts=True,
        update_fields=["origin_url", "visit_type"],
        unique_fields=["origin_url", "visit_type"],
    )

    # associate origins with request
    SaveBulkOrigin.requests.through.objects.bulk_create(
        [
            SaveBulkOrigin.requests.through(
                savebulkorigin_id=save_bulk_origin.id,
                savebulkrequest_id=save_bulk_request.id,
            )
            for save_bulk_origin in save_bulk_origins
        ]
    )

    return str(save_bulk_request.id)


def _rejected_response(
    error_reason: str,
    status: int = 400,
    rejected_origins: Optional[List[Dict[str, str]]] = None,
) -> Response:
    resp_data: Dict[str, Any] = {
        "status": "rejected",
        "reason": error_reason,
    }
    if rejected_origins:
        resp_data["rejected_origins"] = rejected_origins
    return Response(
        resp_data,
        status=status,
    )


class OriginsDataCSVParser(BaseParser):
    media_type = "text/csv"

    def parse(self, stream, media_type=None, parser_context=None):
        try:
            reader = csv.DictReader(
                io.StringIO(stream.read().decode()),
                fieldnames=["origin_url", "visit_type"],
            )
            return [row for row in reader]
        except csv.Error as e:
            raise ParseError(f"CSV data failed to be parsed: {force_str(e)}.")


@api_doc("/origin/save/bulk/", category="Request archival")
@format_docstring()
@api_route(
    r"/origin/save/bulk/",
    "api-1-save-origin-bulk",
    methods=["POST"],
    never_cache=True,
    api_urls=save_bulk_api_urls,
)
@parser_classes([OriginsDataCSVParser, JSONParser, YAMLParser])
def api_origin_save_bulk(request: Request) -> Response:
    """
    .. http:post:: /api/1/origin/save/bulk/

        Request the saving of multiple software origins into the archive.

        That endpoint enables to request the archival of multiple software origins through
        a POST request containing a list of origin URLs and their visit types in its body.

        The following visit types are supported: ``bzr``, ``cvs``, ``hg``, ``git``,
        ``svn`` and ``tarball-directory``.

        The origins list data can be provided using the following content types:

        - ``text/csv`` (default)

          When using CSV format, first column must contain origin URLs and second
          column the visit types.

          .. code-block::

              "https://git.example.org/user/project","git"
              "https://download.example.org/project/source.tar.gz","tarball-directory"

          To post the content of such file to the endpoint, you can use the following
          ``curl`` command.

          .. code-block:: shell

            $ curl -X POST -H "Authorization: Bearer ****" \\
                -H "Content-Type: text/csv" \\
                --data-binary @/path/to/origins.csv \\
                https://archive.softwareheritage.org/api/1/origin/save/bulk/


        - ``application/json``

          When using JSON format, the following schema must be used.

          .. code-block:: json

              [
                  {{
                      "origin_url": "https://git.example.org/user/project",
                      "visit_type": "git"
                  }},
                  {{
                      "origin_url": "https://download.example.org/project/source.tar.gz",
                      "visit_type": "tarball-directory"
                  }}
              ]

          To post the content of such file to the endpoint, you can use the following
          ``curl`` command.

          .. code-block:: shell

            $ curl -X POST -H "Authorization: Bearer ****" \\
                -H "Content-Type: application/json" \\
                --data-binary @/path/to/origins.json \\
                https://archive.softwareheritage.org/api/1/origin/save/bulk/

        - ``application/yaml``

          When using YAML format, the following schema must be used.

          .. code-block:: yaml

              - origin_url: https://git.example.org/user/project
                visit_type: git

              - origin_url: https://download.example.org/project/source.tar.gz
                visit_type: tarball-directory

          To post the content of such file to the endpoint, you can use the following
          ``curl`` command.

          .. code-block:: shell

            $ curl -X POST -H "Authorization: Bearer ****" \\
                -H "Content-Type: application/yaml" \\
                --data-binary @/path/to/origins.yaml \\
                https://archive.softwareheritage.org/api/1/origin/save/bulk/


        Once received, origins data are checked for correctness by validating URLs and
        verifying if visit types are supported. A request cannot be accepted if at least
        one origin is not valid. All origins with invalid format will be reported in the
        rejected request response.

        .. warning::
            That endpoint is not publicly available and requires authentication and
            special user permission in order to request it.

        {common_headers}

        :reqheader Content-Type: the content type of posted data,
            either ``text/csv`` (default), ``application/json`` or ``application/yaml``

        :>json string status: either ``accepted`` or ``rejected``
        :>json string reason: details about why a request got rejected
        :>json string request_id: request identifier (only when it its accepted)
        :>json array rejected_origins: list of rejected origins and details about the reasons
            (only when the request is rejected)

        :statuscode 200: no error
        :statuscode 400: provided origins data are not valid
        :statuscode 401: request is not authenticated
        :statuscode 403: user does not have permission to query the endpoint
        :statuscode 415: payload format is not supported
    """  # noqa
    # authentication and permission checks
    if not bool(request.user and request.user.is_authenticated):
        return _rejected_response(
            "This API endpoint requires authentication.", status=401
        )
    if not request.user.has_perm(API_SAVE_BULK_PERMISSION):
        return _rejected_response(
            "This API endpoint requires a special user permission.", status=403
        )

    # request data basic checks
    if not request.body:
        return _rejected_response("No origins data were provided in POST request body.")
    if not isinstance(request.data, list):
        return _rejected_response("Origins data must be a list of dict.")

    # check origin URLs are well formed and visit types are supported
    rejected_origins = []
    origins = set()
    for origin in sorted(request.data, key=lambda d: d.get("origin_url", "")):
        origin_url = origin.get("origin_url")
        visit_type = origin.get("visit_type")
        origin_data = (origin_url, visit_type)
        if origin_data in origins:
            continue
        origins.add(origin_data)
        if (not origin_url or not visit_type) or not (
            isinstance(origin_url, str) and isinstance(visit_type, str)
        ):
            rejected_origins.append(
                {
                    "origin": origin,
                    "rejection_reason": (
                        "Provided origin data are malformed, please check provided values."
                    ),
                }
            )
        else:
            try:
                validate_origin_url(origin_url)
            except BadInputExc as e:
                rejected_origins.append(
                    {
                        "origin": origin,
                        "rejection_reason": force_str(e),
                    }
                )
            else:
                if visit_type not in SUPPORTED_VISIT_TYPES:
                    rejected_origins.append(
                        {
                            "origin": origin,
                            "rejection_reason": f"Visit type '{visit_type}' is not supported.",
                        }
                    )

    if rejected_origins:
        return _rejected_response(
            "Some origins data could not be validated.",
            rejected_origins=rejected_origins,
        )

    # register origins data to swh-web database
    save_bulk_request_id = _register_request_and_origins_in_db(
        user_id=str(request.user.id), origins=origins
    )

    # generate URL to be queried by the save-bulk lister
    origins_list_url = reverse(
        "save-origin-bulk-origins-list",
        url_args={"request_id": save_bulk_request_id},
        request=request,
    )
    if get_config().get("instance_name", "").endswith("docker.softwareheritage.org"):
        # modify URL if executed in SWH docker environment to ensure lister can query it
        origins_list_url = (
            urlparse(origins_list_url)._replace(scheme="http", netloc="nginx").geturl()
        )

    # create the save-bulk listing task
    task = create_oneshot_task(
        "list-save-bulk",
        url=origins_list_url,
        instance=save_bulk_request_id,
    )

    scheduler().create_tasks([task])

    return Response(
        {
            "status": "accepted",
            "request_id": save_bulk_request_id,
            "request_info_url": reverse(
                "api-1-save-origin-bulk-request-info",
                url_args={"request_id": save_bulk_request_id},
                request=request,
            ),
        }
    )


class SumbittedOriginInfo(TypedDict):
    origin_url: str
    visit_type: str
    status: str
    last_scheduling_date: Optional[str]
    last_visit_date: Optional[str]
    last_visit_status: Optional[str]
    last_snapshot_swhid: Optional[str]
    rejection_reason: Optional[str]
    browse_url: Optional[str]


@api_doc("/origin/save/bulk/request/", category="Request archival")
@format_docstring()
@api_route(
    "/origin/save/bulk/request/<uuid:request_id>/",
    "api-1-save-origin-bulk-request-info",
    never_cache=True,
    api_urls=save_bulk_api_urls,
)
def api_origin_save_bulk_request_info(request: Request, request_id: UUID):
    """
    .. http:get:: /api/1/origin/save/bulk/request/(request_id)/

        Get feedback about loading statuses of origins submitted through a save bulk request.

        That endpoint enables to track the archival statuses of origins sumitted through a POST
        request using the :http:post:`/api/1/origin/save/bulk/` endpoint. Info about submitted
        origins are returned in a paginated way.

        .. note::
            Only origin visits whose dates are greater than the request date are reported by
            that endpoint.

        .. warning::
            That endpoint is not publicly available and requires authentication and
            special user permission in order to request it. Staff users are also
            allowed to query it.

        .. warning::
            Only the user that created a save bulk request or a staff user can get
            feedback about it.

        :param string request_id: UUID identifier of a save bulk request

        :query number page: The submitted origins info page number to retrieve
        :query number per_page: Number of submitted origins info per page, default to 1000,
            maximum is 10000

        :>jsonarr string origin_url: URL of submitted origin
        :>jsonarr string visit_type: visit type for the origin
        :>jsonarr string status: submitted origin status, either ``pending``, ``accepted``
            or ``rejected``
        :>jsonarr date last_scheduling_date: ISO8601/RFC3339 representation of the last date
            (in UTC) when the origin was scheduled for loading into the archive, ``null`` if
            the origin got rejected
        :>jsonarr date last_visit_date: ISO8601/RFC3339 representation of the last date
            (in UTC) when the origin was visited by Software Heritage, ``null`` if the origin
            got rejected or was not visited yet
        :>jsonarr string last_visit_status: last visit status for the origin, either
            ``successful`` or ``failed``, ``null`` if the origin got rejected or was not
            visited yet
        :>jsonarr string last_snapshot_swhid: last produced snapshot SWHID associated to the
            visit, ``null`` if the origin got rejected or was not visited yet
        :>jsonarr string rejection_reason: if the origin got rejected gives more details
            about it
        :>jsonarr string browse_url: URL to browse the submitted origin if it got accepted
            and loaded into the archive, ``null`` if the origin got rejected or was not
            visited yet

        {common_headers}
        {resheader_link}

        :statuscode 200: no error
        :statuscode 401: request is not authenticated
        :statuscode 403: user does not have permission to query the endpoint or get feedback
            about a request he did not submit
    """
    # authentication and permission checks
    if not bool(request.user and request.user.is_authenticated):
        raise UnauthorizedExc("This API endpoint requires authentication.")
    if (
        not request.user.has_perm(API_SAVE_BULK_PERMISSION)
        and not request.user.is_staff
    ):
        raise ForbiddenExc("This API endpoint requires a special user permission.")

    request_id_str = str(request_id)

    # fetch request info
    try:
        save_bulk_request = SaveBulkRequest.objects.get(id=request_id_str)
    except SaveBulkRequest.DoesNotExist:
        raise NotFoundExc(f"Save bulk request with id {request_id_str} not found!")

    # only the user that created the request can retrieve its detailed info
    if save_bulk_request.user_id != str(request.user.id) and not request.user.is_staff:
        raise ForbiddenExc(
            f"Save bulk request with id {request_id_str} was not created with "
            "your user account!"
        )

    # get the lister associated to the request
    lister = scheduler().get_lister("save-bulk", instance_name=request_id_str)

    # get list of origins rejected by the lister
    lister_state = lister.current_state if lister else {}
    rejected_origins = {
        (rejected_origin["origin_url"], rejected_origin["visit_type"]): rejected_origin
        for rejected_origin in lister_state.get("rejected_origins", [])
    }

    # fetch the page of submitted origins to get loadings info
    page_num = int(request.GET.get("page", 1))
    per_page = int(request.GET.get("per_page", 1000))
    per_page = min(per_page, 10000)

    submitted_origins = SaveBulkOrigin.objects.filter(
        requests__in=[save_bulk_request]
    ).order_by("origin_url")

    paginator = Paginator(submitted_origins, per_page)
    try:
        page = paginator.page(page_num)
    except EmptyPage:
        return []

    # fetch listed origins filtered by URLs
    listed_origins = {}
    if lister:
        listed_origins = {
            (listed_origin.url, listed_origin.visit_type): listed_origin
            for listed_origin in scheduler()
            .get_listed_origins(
                lister.id,
                urls=[origin.origin_url for origin in page.object_list],
                limit=per_page,
            )
            .results
        }

    # get origin visit statistics from scheduler
    origin_visit_stats = {
        (visit_stats.url, visit_stats.visit_type): visit_stats
        for visit_stats in scheduler().origin_visit_stats_get(
            (origin.origin_url, origin.visit_type) for origin in page.object_list
        )
    }

    # build response
    response_data = []
    for origin in page.object_list:
        origin_key = (origin.origin_url, origin.visit_type)
        status = "pending"
        last_scheduled = None
        last_visit_date = None
        last_visit_status = None
        last_snapshot = None
        rejection_reason = None
        browse_url = None
        if origin_key in rejected_origins:
            # origin rejected by lister, add rejection reason
            status = "rejected"
            rejection_reason = rejected_origins[origin_key]["reason"]
        elif origin_key in listed_origins:
            # origin accepted by lister, get origin visit stats
            status = "accepted"
            if origin_key in origin_visit_stats:
                last_scheduled = origin_visit_stats[origin_key].last_scheduled
                last_visit_date = origin_visit_stats[origin_key].last_visit
                last_visit_status = origin_visit_stats[origin_key].last_visit_status
                last_snapshot = origin_visit_stats[origin_key].last_snapshot
            if last_snapshot:
                browse_url = reverse(
                    "browse-swhid",
                    url_args={
                        "swhid": QualifiedSWHID(
                            object_type=ObjectType.SNAPSHOT,
                            object_id=last_snapshot,
                            origin=origin.origin_url,
                        )
                    },
                    request=request,
                )
            if last_scheduled and last_scheduled > save_bulk_request.request_date:
                # only report visit date greater than request date
                if last_visit_date and last_visit_date < save_bulk_request.request_date:
                    last_visit_date = None
                    last_visit_status = None
                    last_snapshot = None
                    browse_url = None
        # add submitted origin info to response data
        response_data.append(
            SumbittedOriginInfo(
                origin_url=origin.origin_url,
                visit_type=origin.visit_type,
                status=status,
                last_scheduling_date=(
                    datetime_to_utc(last_scheduled).isoformat()
                    if last_scheduled
                    else None
                ),
                last_visit_date=(
                    datetime_to_utc(last_visit_date).isoformat()
                    if last_visit_date
                    else None
                ),
                last_visit_status=(
                    last_visit_status.value if last_visit_status else None
                ),
                last_snapshot_swhid=(
                    str(
                        CoreSWHID(
                            object_type=ObjectType.SNAPSHOT, object_id=last_snapshot
                        )
                    )
                    if last_snapshot
                    else None
                ),
                rejection_reason=rejection_reason,
                browse_url=browse_url,
            )
        )

    response: Dict[str, Any] = {"results": response_data, "headers": {}}

    # compute link header for pagination
    if page.has_previous():
        response["headers"]["link-prev"] = reverse(
            "api-1-save-origin-bulk-request-info",
            url_args={"request_id": request_id},
            query_params={"per_page": str(per_page), "page": str(page_num - 1)},
            request=request,
        )

    if page.has_next():
        response["headers"]["link-next"] = reverse(
            "api-1-save-origin-bulk-request-info",
            url_args={"request_id": request_id},
            query_params={"per_page": str(per_page), "page": str(page_num + 1)},
            request=request,
        )

    return response


@api_doc("/origin/save/bulk/requests/", category="Request archival")
@format_docstring()
@api_route(
    "/origin/save/bulk/requests/",
    "api-1-save-origin-bulk-requests",
    never_cache=True,
    api_urls=save_bulk_api_urls,
)
def api_origin_save_bulk_requests(request: Request):
    """
    .. http:get:: /api/1/origin/save/bulk/requests/

        List previously submitted save bulk requests.

        That endpoint enables to list the save bulk requests submitted by
        your user account and get their info URLs
        (see :http:get:`/api/1/origin/save/bulk/request/(request_id)/`).
        That list is returned in a paginated way if the number or requests
        is large.

        .. warning::
            That endpoint is not publicly available and requires authentication and
            special user permission in order to request it.

        :query number page: The submitted requests page number to retrieve
        :query number per_page: Number of submitted requests per page, default to 1000,
            maximum is 10000

        :>jsonarr string request_id: UUID identifier of the request
        :>jsonarr date request_date: the date the request was submitted
        :>jsonarr string request_info_url: URL to get detailed info about the request

        {common_headers}
        {resheader_link}

        :statuscode 200: no error
        :statuscode 401: request is not authenticated
        :statuscode 403: user does not have permission to query the endpoint

    """

    if not bool(request.user and request.user.is_authenticated):
        raise UnauthorizedExc("This API endpoint requires authentication.")
    if not request.user.has_perm(API_SAVE_BULK_PERMISSION):
        raise ForbiddenExc("This API endpoint requires a special user permission.")

    save_bulk_requests = SaveBulkRequest.objects.filter(
        user_id=str(request.user.id)
    ).order_by("-request_date")

    page_num = int(request.GET.get("page", 1))
    per_page = int(request.GET.get("per_page", 1000))
    per_page = min(per_page, 10000)

    paginator = Paginator(save_bulk_requests, per_page)
    try:
        page = paginator.page(page_num)
    except EmptyPage:
        return []

    response_data = [
        {
            "request_id": str(save_bulk_request.id),
            "request_date": save_bulk_request.request_date.isoformat(),
            "request_info_url": reverse(
                "api-1-save-origin-bulk-request-info",
                url_args={"request_id": str(save_bulk_request.id)},
                request=request,
            ),
        }
        for save_bulk_request in page.object_list
    ]

    response: Dict[str, Any] = {"results": response_data, "headers": {}}

    if page.has_previous():
        response["headers"]["link-prev"] = reverse(
            "api-1-save-origin-bulk-requests",
            query_params={"per_page": str(per_page), "page": str(page_num - 1)},
            request=request,
        )

    if page.has_next():
        response["headers"]["link-next"] = reverse(
            "api-1-save-origin-bulk-requests",
            query_params={"per_page": str(per_page), "page": str(page_num + 1)},
            request=request,
        )

    return response



# Copyright (C) 2022-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


import email
import email.message
import email.policy
import logging
from typing import Callable, Optional, Type

from swh.web.inbound_email import signals
from swh.web.utils.exc import sentry_capture_exception

logger = logging.getLogger(__name__)


class MessageHandler:
    def __init__(self, raw_message: bytes, sender: Optional[Type] = None):
        try:
            self.message = email.message_from_bytes(
                raw_message, policy=email.policy.default
            )
        except Exception as exc:
            sentry_capture_exception(exc)
            self.failed_message(raw_message)
            # XXX make sure having logging doesn't make postfix unhappy
            logger.exception("Could not convert email from bytes")
            return

        self.sender = sender or self.__class__

    def handle(self):
        responses = signals.email_received.send_robust(
            sender=self.sender, message=self.message
        )

        handled = False
        for receiver, response in responses:
            if isinstance(response, Exception):
                sentry_capture_exception(response)
                self.failing_receiver(receiver)
                logger.error(
                    "Receiver produced the following exception", exc_info=response
                )
            elif response is signals.EmailProcessingStatus.FAILED:
                self.failing_receiver(receiver)
            elif response is signals.EmailProcessingStatus.PROCESSED:
                handled = True

        if not handled:
            self.unhandled_message()

        return handled

    @classmethod
    def failed_message(cls, raw_message: bytes):
        # TODO: forward email as attachment for inspection
        logger.error("Failed message: %s", raw_message.decode("ascii", "replace"))

    def failing_receiver(self, receiver: Callable):
        # TODO: forward email for inspection
        logger.error(
            "Failed receiver %s:%s; message: %s",
            receiver.__module__,
            receiver.__qualname__,
            str(self.message),
        )

    def unhandled_message(self):
        # TODO: pass email through to a fallback alias?
        logger.error("Unhandled message: %s", str(self.message))



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING, Any, Union
import uuid

from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import validate_domain_name, validate_email
from django.db import IntegrityError, models, transaction
from django.db.models import Q
from django.db.models.query import QuerySet
from django.urls import reverse
from django.utils import timezone
from django.utils.crypto import get_random_string
from django.utils.translation import gettext as _

from .apps import APP_LABEL

if TYPE_CHECKING:
    from django.contrib.sessions.backends.base import SessionBase


class BaseModel(models.Model):
    """An abstract base model to provide UUID pks and timestamps."""

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    """UUID primary key"""

    created_at = models.DateTimeField(_("created"), auto_now_add=True)
    """Creation date"""

    updated_at = models.DateTimeField(_("updated"), auto_now=True)
    """Last update"""

    class Meta:
        abstract = True


class OriginOutcome(models.TextChoices):
    VALIDATING = "validating", _("Validating")
    REJECTED = "rejected", _("Rejected")
    MAILMAP = "mailmap", _("Mailmap")
    MASK = "mask", _("Permanent mask")
    TAKEDOWN = "takedown", _("Takedown")
    BLOCK = "block", _("Takedown and block")


class OriginOwnership(models.TextChoices):
    UNKNOWN = "unknown", _("?")
    OWNER = "owner", _("Requester is the owner of the origin")
    FORK = "fork", _("Fork of an origin owned by the requester")
    OTHER = "other", _("Origin has no direct link with the requester")


class Origin(BaseModel):
    """Origins concerned by an `Alteration`."""

    url = models.URLField(_("URL"))
    """Origin's URL"""

    alteration = models.ForeignKey(
        "Alteration",
        related_name="origins",
        on_delete=models.RESTRICT,
        null=False,
    )
    """Alteration FK"""

    outcome = models.CharField(
        _("outcome"),
        max_length=20,
        choices=OriginOutcome,
        default=OriginOutcome.VALIDATING,
    )
    """Outcome for this origin"""

    reason = models.TextField(_("reason for this outcome"), blank=True)
    """Outcome's reason"""

    code_license = models.TextField(_("license found in code"), blank=True)
    """License found in / guessed from the source code"""

    available = models.BooleanField(
        _("available"),
        help_text=_("Is this URL is still available online ?"),
        null=True,
    )
    """URL is still available"""

    ownership = models.CharField(
        _("owner"),
        help_text=_("Is Requester the owner of this origin or is it a fork ?"),
        max_length=20,
        choices=OriginOwnership,
        default=OriginOwnership.UNKNOWN,
    )
    """Is Requester the owner of this origin or is it a fork"""

    class Meta:
        app_label = APP_LABEL
        db_table = "origin"
        indexes = [models.Index(fields=["url"])]
        constraints = [
            models.UniqueConstraint(fields=["alteration_id", "url"], name="unique_url"),
        ]

    def __repr__(self) -> str:
        return f"<Origin: {self.url} ({self.pk})>"

    def __str__(self) -> str:
        return self.url

    def get_admin_url(self):
        return reverse(
            "alteration-origin-admin",
            kwargs={"pk": self.pk, "alteration_pk": self.alteration_id},
        )


class AlterationManager(models.Manager["Alteration"]):
    """Custom ``Alteration`` manager."""

    def search(self, query: str) -> QuerySet[Alteration, Alteration]:
        """A basic search for requests.

        Will find requests where `query` is to be found (case insensitive) in either:
        * its ``reasons`` field
        * its ``expected_outcome`` field
        * one of its ``Origin`` url
        * Requester's email

        Args:
            query: search query
        """
        base_queryset = super().get_queryset()
        search_filters = (
            Q(reasons__icontains=query)
            | Q(expected_outcome__icontains=query)
            | Q(origins__url__icontains=query)
            | Q(email__icontains=query)
        )
        return base_queryset.filter(search_filters).distinct()


class AlterationStatus(models.TextChoices):
    VALIDATING = "validating", _("Validating")
    PLANNING = "planning", _("Planning")
    EXECUTING = "executing", _("Executing")
    PROCESSED = "processed", _("Processed")
    REJECTED = "rejected", _("Rejected")
    CLOSED = "closed", _("Closed")
    ARCHIVED = "archived", _("Archived")

    __empty__ = _("Filter by status")


class AlterationCategory(models.TextChoices):
    COPYRIGHT = "copyright", _("Copyright / License infringement")
    PII = "pii", _("Personal Identifiable Information")
    LEGAL = "legal", _("Other legal matters")

    __empty__ = _("Filter by category")


class Alteration(BaseModel):
    """An alteration request."""

    status = models.CharField(
        _("status"),
        max_length=20,
        choices=AlterationStatus,
        default=AlterationStatus.VALIDATING,
    )
    """Progression indicator"""

    category = models.CharField(
        _("category"), max_length=20, choices=AlterationCategory
    )
    """The category/type of this alteration"""

    reasons = models.TextField(_("reasons"), null=False)
    """The alteration request reasons"""

    expected_outcome = models.TextField(_("expected outcome"), null=False)
    """The Requester expectations"""

    email = models.EmailField(_("requester's email"))
    """Requester's email"""

    objects = AlterationManager()

    class Meta:
        app_label = APP_LABEL
        db_table = "alteration"
        ordering = ["-created_at"]

    def get_absolute_url(self):
        return reverse("alteration-details", kwargs={"pk": self.pk})

    def get_admin_url(self):
        return reverse("alteration-admin", kwargs={"pk": self.pk})

    @classmethod
    def create_from_assistant(cls, session: SessionBase) -> Alteration:
        """Create an alteration request from the alteration request assistant.

        Email notifications are sent to the requester & operators.

        Args:
            session: django's session store

        Returns:
            an Alteration instance

        Raises:
            DatabaseError: something went wrong while creating objects in the database
        """
        with transaction.atomic():
            alteration = Alteration.objects.create(
                category=session["alteration_category"],
                reasons=session["alteration_reasons"],
                expected_outcome=session["alteration_expected_outcome"],
                email=session["alteration_email"],
            )
            Origin.objects.bulk_create(
                [
                    Origin(alteration=alteration, url=url)
                    for url in session["alteration_origins"]
                ]
            )
            Event.objects.create(
                alteration=alteration,
                category=EventCategory.LOG,
                author=_("Requester"),
                content=_("Alteration request created."),
                internal=False,
            )
        return alteration

    @property
    def is_read_only(self) -> bool:
        return self.status == AlterationStatus.ARCHIVED

    def __repr__(self) -> str:
        return f"<Alteration: {self.category} [{self.email}] ({self.pk})>"

    def __str__(self) -> str:
        return f"{self.get_category_display()} by {self.email}"


class EventRecipient(models.TextChoices):
    REQUESTER = "requester", _("Requester")
    SUPPORT = "support", _("Support")
    MANAGER = "manager", _("Manager")
    LEGAL = "legal", _("Legal")
    TECHNICAL = "technical", _("Technical")


class EventCategory(models.TextChoices):
    LOG = "log", _("Event")
    MESSAGE = "message", _("Message")


class EventManager(models.Manager["Event"]):
    """Custom ``Event`` manager."""

    def get_queryset(self) -> QuerySet[Event, Event]:
        """Filters internal events.

        Returns:
            An ``Event`` queryset with ``internal`` events filtered out.
        """
        return super().get_queryset().filter(internal=False)


class Event(BaseModel):
    """An event on an `Alteration`.

    An event could be
    * a log of a status change or a modification of any another field / Origin
    * a message between recipients
    """

    alteration = models.ForeignKey(
        "Alteration",
        related_name="events",
        on_delete=models.RESTRICT,
        null=False,
    )
    """Alteration FK"""

    category = models.CharField(_("category"), max_length=20, choices=EventCategory)
    """Category/type of event"""

    author = models.CharField(_("author"), max_length=200, blank=True)
    """The name of the author of this event"""

    recipient = models.CharField(
        _("recipient role"), max_length=20, choices=EventRecipient, blank=True
    )
    """The role targeted by this event, a value is required to send notifications"""

    content = models.TextField(_("content"), blank=True)
    """The event's textual content"""

    internal = models.BooleanField(
        _("internal"),
        default=True,
        help_text=_(
            "Internal messages are not visible in the Requester activity log to avoid "
            "unnecessary noise, they must not be used for confidential exchanges "
            "between team members and could be requested by the user."
        ),
    )
    """Internal actions are not visible to the Requester"""

    objects = models.Manager()
    public_objects = EventManager()

    class Meta:
        app_label = APP_LABEL
        db_table = "event"
        ordering = ["-created_at"]

    def __repr__(self) -> str:
        return f"<Event: {self.category} [{self.author}] ({self.pk})>"

    def __str__(self) -> str:
        return f"{self.get_category_display()} by {self.author} for {self.alteration}"

    def get_admin_url(self):
        return reverse(
            "alteration-event-admin",
            kwargs={"pk": self.pk, "alteration_pk": self.alteration_id},
        )


TOKEN_TTL = 15 * 60
"""Token expiration delay in seconds"""

TOKEN_NBYTES = 20
"""Token length, must be less than ~1.3 x `Token.value` max length"""


def _default_token_expires_at():
    return timezone.now() + timedelta(seconds=TOKEN_TTL)


def _default_token_value():
    return get_random_string(TOKEN_NBYTES)


class Token(BaseModel):
    """Ephemeral auth tokens to access an `Alteration` or validate an email."""

    alteration = models.ForeignKey(
        "Alteration",
        related_name="tokens",
        on_delete=models.CASCADE,
        null=True,
    )
    """Alteration FK"""

    email = models.EmailField(_("email"), null=True)
    """An email address to validate"""

    value = models.CharField(_("value"), max_length=32, default=_default_token_value)
    """Token value"""

    expires_at = models.DateTimeField(
        _("expiration date"), null=False, default=_default_token_expires_at
    )
    """Token expiration date"""

    objects = models.Manager()

    class Meta:
        app_label = APP_LABEL
        db_table = "token"
        indexes = [models.Index(fields=["value"]), models.Index(fields=["email"])]
        constraints = [
            models.UniqueConstraint(fields=["value"], name="unique_token_value"),
        ]

    def __repr__(self) -> str:
        return f"<Token: {self.value}>"

    def __str__(self) -> str:
        item = self.alteration if self.alteration else self.email
        return f"{self.value} for {item}"

    def get_absolute_url(self) -> str:
        route = (
            "alteration-link" if self.alteration else "alteration-email-verification"
        )
        return reverse(route, kwargs={"value": self.value})

    @property
    def expired(self) -> bool:
        return self.expires_at < timezone.now()

    @classmethod
    def create_for(cls, obj: Union[Alteration, str]) -> Token:
        """Create a token for an `Alteration` request or an email verification.

        Args:
            obj: an `Alteration` instance or an email

        Returns:
            A `Token` instance

        Raises:
            ValueError: `obj` is neither an `Alteration` nor an email
            IntegrityError: if we're not able to create a token after 5 attempts
        """
        params: dict[str, Union[Alteration, str]]
        if isinstance(obj, Alteration):
            params = {"alteration": obj}
        elif isinstance(obj, str):
            params = {"email": obj}
        else:
            raise ValueError(
                _(
                    "Invalid parameter %(obj)s, a token can only be created for "
                    "an alteration request or an email address."
                )
                % {"obj": obj}
            )

        for attempt in range(1, 5):
            try:
                return cls.objects.create(**params)
            except IntegrityError as exc:
                last_exception = exc
        raise IntegrityError(
            f"Could not create a unique token after {attempt} attempts"
        ) from last_exception


def validate_email_or_domain(value: Any) -> None:
    """Check if value is a valid email or a domain name.

    Args:
        value: a string

    Raises:
        ValidationError: `value` is neither an email address nor a domain name

    Returns:
        True is `value` is an email or a domain
    """
    try:
        validate_email(value)
        return
    except ValidationError:
        pass
    try:
        validate_domain_name(value)
        return
    except ValidationError:
        pass

    raise ValidationError(
        _("%(value)s is neither an email address nor a domain name."),
        params={"value": value},
    )


class LowerCharField(models.CharField):
    """A lowercased `CharField`."""

    def pre_save(self, model_instance: models.Model, add: bool) -> Any:
        """Lower case value before saving the instance."""
        value = getattr(model_instance, self.attname)
        if isinstance(value, str):
            setattr(model_instance, self.attname, value.lower())
        return super().pre_save(model_instance, add)


class BlockList(BaseModel):
    """Email or domain block list.

    Email addresses or domains found in this table are not allowed to initiate an
    `Alteration` request.
    """

    email_or_domain = LowerCharField(
        _("email or domain"), max_length=254, validators=[validate_email_or_domain]
    )
    """An email address or a domain"""

    reason = models.TextField(_("reasons"), blank=True)
    """Reason for the ban (ie. added by an admin or requested by a user)"""

    class Meta:
        app_label = APP_LABEL
        db_table = "block_list"
        indexes = [models.Index(fields=["email_or_domain"])]
        constraints = [
            models.UniqueConstraint(
                fields=["email_or_domain"], name="unique_email_or_domain"
            ),
        ]

    @classmethod
    def is_blocked(cls, email: str) -> bool:
        """Check if `email` or its domain is blocked.

        We use email or domains stored in this table and optionally domains from
        https://github.com/disposable-email-domains/disposable-email-domains

        Args:
            value: an email address

        Returns:
            True if the address is blocked

        Raises:
            ValueError: `email` is an invalid email address
        """
        try:
            validate_email(email)
        except ValidationError as exc:
            raise ValueError(_("Invalid email value")) from exc
        email = email.lower()
        domain = email.split("@")[-1]

        if settings.ALTER_SETTINGS.get("block_disposable_email_domains"):
            from disposable_email_domains import blocklist

            disposable = domain in blocklist
        else:
            disposable = False

        return (
            disposable
            or cls.objects.filter(email_or_domain__in=[email, domain]).exists()
        )



# Copyright (C) 2022-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import abc
from typing import Any, Dict, Tuple

from rest_framework.request import Request

from swh.web.api.apidoc import api_doc
from swh.web.api.apiurls import APIUrls, api_route
from swh.web.save_code_now.origin_save import create_save_origin_request
from swh.web.utils import reverse
from swh.web.utils.exc import BadInputExc

webhooks_api_urls = APIUrls()


SUPPORTED_FORGE_TYPES = set()


class OriginSaveWebhookReceiver(abc.ABC):
    FORGE_TYPE: str
    WEBHOOK_GUIDE_URL: str
    REPO_TYPES: str

    @abc.abstractmethod
    def is_forge_request(self, request: Request) -> bool: ...

    def is_ping_event(self, request: Request) -> bool:
        return False

    @abc.abstractmethod
    def is_push_event(self, request: Request) -> bool: ...

    @abc.abstractmethod
    def extract_repo_info(self, request: Request) -> Tuple[str, str, bool]:
        """Extract and return a tuple (repository_url, visit_type, private) from
        the forge webhook payload."""
        ...

    def __init__(self):
        self.__doc__ = f"""
        .. http:post:: /api/1/origin/save/webhook/{self.FORGE_TYPE.lower()}/

            Webhook receiver for {self.FORGE_TYPE} to request or update the archival of
            a repository when new commits are pushed to it.

            To add such webhook to one of your {self.REPO_TYPES} repository hosted on
            {self.FORGE_TYPE}, please follow `{self.FORGE_TYPE}'s webhooks guide
            <{self.WEBHOOK_GUIDE_URL}>`_.

            The expected content type for the webhook payload must be ``application/json``.

            Please note that to avoid abusing the archival service offered by Software Heritage
            at most **one request per hour is created** so the effective loading of the
            repository into the archive might be delayed.

            :>json number id: the save request identifier
            :>json string request_url: Web API URL to follow up on that request
            :>json string origin_url: the url of the origin to save
            :>json string visit_type: the type of visit to perform
            :>json string save_request_date: the date (in iso format) the save
                request was issued
            :>json string save_request_status: the status of the save request,
                either **accepted**, **rejected** or **pending**
            :>json string save_task_status: the status of the origin saving task,
                either **not created**, **pending**, **scheduled**, **running**,
                **succeeded** or **failed**
            :>json string save_task_next_run: the date and time from which the
                request is executed

            :statuscode 200: save request for repository has been successfully created
                from the webhook payload.
            :statuscode 400: no save request has been created due to invalid POST
                request or missing data in webhook payload
        """
        self.__name__ = "api_origin_save_webhook_{self.FORGE_TYPE.lower()}"
        SUPPORTED_FORGE_TYPES.add(self.FORGE_TYPE.lower())
        api_doc(
            f"/origin/save/webhook/{self.FORGE_TYPE.lower()}/",
            category="Request archival",
        )(self)
        api_route(
            f"/origin/save/webhook/{self.FORGE_TYPE.lower()}/",
            f"api-1-origin-save-webhook-{self.FORGE_TYPE.lower()}",
            methods=["POST"],
            api_urls=webhooks_api_urls,
        )(self)

    def __call__(
        self,
        request: Request,
    ) -> Dict[str, Any]:
        if not self.is_forge_request(request):
            raise BadInputExc(
                f"POST request was not sent by a {self.FORGE_TYPE} webhook and "
                "has not been processed."
            )

        if self.is_ping_event(request):
            return {"message": "pong"}

        if not self.is_push_event(request):
            raise BadInputExc(
                f"Event sent by {self.FORGE_TYPE} webhook is not a push one, request "
                "has not been processed."
            )

        content_type = request.headers.get("Content-Type")
        if content_type and not content_type.startswith("application/json"):
            raise BadInputExc(
                f"Invalid content type '{content_type}' for the POST request sent by "
                f"{self.FORGE_TYPE} webhook, it should be 'application/json'."
            )

        repo_url, visit_type, private = self.extract_repo_info(request)
        if not repo_url:
            raise BadInputExc(
                f"Repository URL could not be extracted from {self.FORGE_TYPE} webhook "
                f"payload."
            )
        if not visit_type:
            raise BadInputExc(
                f"Visit type could not be determined for repository {repo_url}."
            )
        if private:
            raise BadInputExc(
                f"Repository {repo_url} is private and cannot be cloned without authentication."
            )

        save_request = create_save_origin_request(
            visit_type=visit_type,
            origin_url=repo_url,
            from_webhook=True,
            webhook_origin=self.FORGE_TYPE.lower(),
        )

        assert save_request["next_run"] is not None

        return {
            "id": save_request["id"],
            "request_url": reverse(
                "api-1-save-origin",
                url_args={"request_id": save_request["id"]},
                request=request,
            ),
            "origin_url": save_request["origin_url"],
            "visit_type": save_request["visit_type"],
            "save_request_date": save_request["save_request_date"],
            "save_request_status": save_request["save_request_status"],
            "save_task_status": save_request["save_task_status"],
            "save_task_next_run": save_request["next_run"].isoformat(),
        }



# Copyright (C) 2017-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import List, Optional

# call this early to ensure all the 'swhid' url path converted is registered
# so any url registration will work as intended
import swh.web.utils.url_path_converters  # noqa: F401
from swh.web.utils.urlsindex import UrlsIndex

browse_urls = UrlsIndex()


def browse_route(
    *url_patterns: str,
    view_name: Optional[str] = None,
    checksum_args: Optional[List[str]] = None,
):
    """
    Decorator to ease the registration of a swh-web browse endpoint

    Args:
        url_patterns: list of url patterns used by Django to identify the
            browse routes
        view_name: the name of the Django view associated to the routes used
            to reverse the url
    """
    url_patterns = tuple("browse/" + url_pattern for url_pattern in url_patterns)
    view_name = view_name

    def decorator(f):
        # register the route and its view in the browse endpoints index
        for url_pattern in url_patterns:
            browse_urls.add_url_pattern(url_pattern, f, view_name)

        if checksum_args:
            browse_urls.add_redirect_for_checksum_args(
                view_name, url_patterns, checksum_args
            )

        return f

    return decorator



# Copyright (C) 2022-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
from typing import Any, Dict

from django.contrib.auth.decorators import permission_required
from django.core.paginator import Paginator
from django.db import IntegrityError
from django.db.models import Q
from django.http.request import HttpRequest
from django.http.response import (
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseNotFound,
    JsonResponse,
)
from django.shortcuts import render
from rest_framework import serializers
from rest_framework.decorators import api_view
from rest_framework.request import Request
from rest_framework.response import Response

from swh.web.auth.utils import (
    MAILMAP_ADMIN_PERMISSION,
    MAILMAP_PERMISSION,
    any_permission_required,
)
from swh.web.mailmap.models import UserMailmap, UserMailmapEvent
from swh.web.utils import datatables_order_params, datatables_pagination_params


class UserMailmapSerializer(serializers.ModelSerializer):
    class Meta:
        model = UserMailmap
        fields = "__all__"


@api_view(["GET"])
@any_permission_required(MAILMAP_PERMISSION, MAILMAP_ADMIN_PERMISSION)
def profile_list_mailmap(request: Request) -> HttpResponse:
    mailmap_admin = request.user.has_perm(MAILMAP_ADMIN_PERMISSION)

    mms = UserMailmap.objects.filter(
        user_id=None if mailmap_admin else str(request.user.id)
    ).all()
    return Response(UserMailmapSerializer(mms, many=True).data)


@api_view(["POST"])
@any_permission_required(MAILMAP_PERMISSION, MAILMAP_ADMIN_PERMISSION)
def profile_add_mailmap(request: Request) -> HttpResponse:
    mailmap_admin = request.user.has_perm(MAILMAP_ADMIN_PERMISSION)

    event = UserMailmapEvent.objects.create(
        user_id=str(request.user.id),
        request_type="add",
        request=json.dumps(request.data),
    )

    from_email = request.data.pop("from_email", None)
    if not from_email:
        return HttpResponseBadRequest(
            "'from_email' must be provided and non-empty.", content_type="text/plain"
        )

    user_id = None if mailmap_admin else str(request.user.id)

    from_email_verified = request.data.pop("from_email_verified", False)
    if mailmap_admin:
        # consider email verified when mailmap is added by admin
        from_email_verified = True

    try:
        UserMailmap.objects.create(
            user_id=user_id,
            from_email=from_email,
            from_email_verified=from_email_verified,
            **request.data,
        )
    except IntegrityError as e:
        if (
            "user_mailmap_from_email_key" in e.args[0]
            or "user_mailmap.from_email" in e.args[0]
        ):
            return HttpResponseBadRequest(
                "This 'from_email' already exists.", content_type="text/plain"
            )
        else:
            raise

    event.successful = True
    event.save()

    mm = UserMailmap.objects.get(user_id=user_id, from_email=from_email)
    return Response(UserMailmapSerializer(mm).data)


@api_view(["POST"])
@any_permission_required(MAILMAP_PERMISSION, MAILMAP_ADMIN_PERMISSION)
def profile_update_mailmap(request: Request) -> HttpResponse:
    mailmap_admin = request.user.has_perm(MAILMAP_ADMIN_PERMISSION)

    event = UserMailmapEvent.objects.create(
        user_id=str(request.user.id),
        request_type="update",
        request=json.dumps(request.data),
    )

    from_email = request.data.pop("from_email", None)
    if not from_email:
        return HttpResponseBadRequest(
            "'from_email' must be provided and non-empty.", content_type="text/plain"
        )

    user_id = None if mailmap_admin else str(request.user.id)

    try:
        to_update = (
            UserMailmap.objects.filter(user_id=user_id)
            .filter(from_email=from_email)
            .get()
        )
    except UserMailmap.DoesNotExist:
        return HttpResponseNotFound("'from_email' cannot be found in mailmaps.")

    for attr, value in request.data.items():
        setattr(to_update, attr, value)

    to_update.save()

    event.successful = True
    event.save()

    mm = UserMailmap.objects.get(user_id=user_id, from_email=from_email)
    return Response(UserMailmapSerializer(mm).data)


@any_permission_required(MAILMAP_PERMISSION, MAILMAP_ADMIN_PERMISSION)
def profile_list_mailmap_datatables(request: HttpRequest) -> HttpResponse:
    mailmap_admin = request.user.has_perm(MAILMAP_ADMIN_PERMISSION)

    mailmaps = UserMailmap.objects.filter(
        user_id=None if mailmap_admin else str(request.user.id)
    )

    search_value = request.GET.get("search[value]", "")

    field_order = datatables_order_params(request, "from_email", "asc")

    mailmaps = mailmaps.order_by(*field_order)

    table_data: Dict[str, Any] = {}
    table_data["draw"] = int(request.GET.get("draw", 1))
    table_data["recordsTotal"] = mailmaps.count()

    length, page = datatables_pagination_params(request)

    if search_value:
        mailmaps = mailmaps.filter(
            Q(from_email__icontains=search_value)
            | Q(display_name__icontains=search_value)
        )

    table_data["recordsFiltered"] = mailmaps.count()

    paginator = Paginator(mailmaps, length)

    mailmaps_data = [
        UserMailmapSerializer(mm).data for mm in paginator.page(int(page)).object_list
    ]

    table_data["data"] = mailmaps_data

    return JsonResponse(table_data)


@permission_required(MAILMAP_ADMIN_PERMISSION)
def admin_mailmap(request):
    return render(request, "admin/mailmap.html")
